use super::{
    get_player_item_or_fail, get_user_or_fail, parsing::parse_command_name,
    rent::recursively_destroy_or_move_item, search_item_for_user, user_error, UResult, UserVerb,
    UserVerbRef, VerbContext,
};
use crate::{
    db::ItemSearchParams,
    language::caps_first,
    models::corp::{CorpCommType, CorpMembership, CorpPermission},
};
use ansi::{ansi, ignore_special_characters};
use async_trait::async_trait;
use chrono::Utc;
use humantime;
use itertools::Itertools;
use std::collections::BTreeSet;

pub fn check_corp_perm(perm: &CorpPermission, mem: &CorpMembership) -> bool {
    mem.permissions
        .iter()
        .any(|p| *p == CorpPermission::Holder || *p == *perm)
}

async fn corp_invite(ctx: &mut VerbContext<'_>, remaining: &str) -> UResult<()> {
    let (target_raw, into_raw) = match remaining.rsplit_once(" into ") {
        None => user_error(
            ansi!("Usage: <bold>corp hire<reset> username <bold>into<reset> corpname").to_owned(),
        )?,
        Some(c) => c,
    };
    let user = get_user_or_fail(ctx)?;
    let player = get_player_item_or_fail(ctx).await?;
    let (corp_id, corp, mem) = match ctx
        .trans
        .match_user_corp_by_name(into_raw.trim(), &user.username)
        .await?
    {
        None => user_error("You don't seem to belong to a matching corp!".to_owned())?,
        Some(c) => c,
    };
    if !check_corp_perm(&CorpPermission::Hire, &mem) || mem.joined_at.is_none() {
        user_error("You don't have hiring permissions for that corp".to_owned())?;
    }

    let target_user = search_item_for_user(
        ctx,
        &ItemSearchParams {
            include_loc_contents: true,
            ..ItemSearchParams::base(&player, target_raw.trim())
        },
    )
    .await?;

    if target_user.item_type != "player" {
        user_error("Only players can be hired.".to_owned())?;
    }

    let (their_sess, _their_sess_dat) = match ctx
        .trans
        .find_session_for_player(&target_user.item_code)
        .await?
    {
        None => user_error("The user needs to be logged in while you hire them.".to_owned())?,
        Some(c) => c,
    };

    if ctx.trans.list_corp_members(&corp_id).await?.len() > 40 {
        user_error(
            "Your corp seems a bit too big to manage already \
                    - fire someone first!"
                .to_owned(),
        )?;
    }

    ctx.trans.expire_old_invites().await?;

    match ctx
        .trans
        .match_user_corp_by_name(into_raw.trim(), &target_user.item_code)
        .await?
    {
        None => (),
        Some((
            _,
            _,
            CorpMembership {
                invited_at: Some(_),
                ..
            },
        )) => user_error(format!(
            "{}'s already been invited.",
            &caps_first(&target_user.pronouns.subject)
        ))?,
        Some((_, _, _)) => user_error(format!(
            "{}'s already hired.",
            &caps_first(&target_user.pronouns.subject)
        ))?,
    };

    let new_mem = CorpMembership {
        invited_at: Some(Utc::now()),
        allow_combat: corp.allow_combat_required,
        permissions: corp.member_permissions.clone(),
        ..Default::default()
    };
    ctx.trans
        .upsert_corp_membership(&corp_id, &target_user.item_code, &new_mem)
        .await?;

    ctx.trans.queue_for_session(&their_sess, Some(&format!(
        ansi!("{} wants to hire you into {}! Type <bold>corp join {}<reset> to accept.{}\n"),
        &player.display_for_sentence(1, true),
        &corp.name,
        &corp.name,
        if new_mem.allow_combat {
            " This corp is configured to allow the leadership to declare war on other corps; if you join, you may be attacked by members of other corps, even without your personal consent."
        } else {
            ""
        }
    ))).await?;
    ctx.trans
        .broadcast_to_corp(
            &corp_id,
            &CorpCommType::Notice,
            None,
            &format!(
                "{} has invited {} to join {}!\n",
                user.username,
                &target_user.display_for_sentence(1, false),
                corp.name
            ),
        )
        .await?;
    Ok(())
}

async fn corp_join(ctx: &mut VerbContext<'_>, remaining: &str) -> UResult<()> {
    let user = get_user_or_fail(ctx)?;
    let (corpid, corp, mut mem) = match ctx
        .trans
        .match_user_corp_by_name(remaining, &user.username)
        .await?
    {
        None => user_error("Corp not found".to_owned())?,
        Some(v) => v,
    };
    if mem.joined_at.is_some() {
        user_error("You have already joined that corp!".to_owned())?;
    }
    mem.joined_at = Some(Utc::now());
    mem.invited_at = None;
    ctx.trans
        .upsert_corp_membership(&corpid, &user.username, &mem)
        .await?;

    ctx.trans
        .broadcast_to_corp(
            &corpid,
            &CorpCommType::Notice,
            None,
            &format!(
                ansi!("There is a loud <red>cheer<reset> as {} accepts an offer to join {}!\n"),
                &user.username, &corp.name
            ),
        )
        .await?;

    Ok(())
}

async fn corp_list(ctx: &mut VerbContext<'_>, _remaining: &str) -> UResult<()> {
    let mut msg = String::new();
    let user = get_user_or_fail(ctx)?;

    let corps = ctx
        .trans
        .get_corp_memberships_for_user(&user.username)
        .await?;
    if corps.is_empty() {
        msg.push_str(ansi!(
            "You don't yet belong to any corps - try <bold>who<reset> to see the \
             corps of people online, and message someone in a corp to see if they \
             will hire you! Or buy a new corp licence at the Kings Office in Melbs.\n"
        ));
    } else {
        msg.push_str("You belong to the following corps:\n");
        msg.push_str(&format!(
            ansi!("<bgblue><white><bold>| {:20} | {:7} | {:5} |<reset>\n"),
            "Name", "Combat?", "Order"
        ));
        for corp in &corps {
            match corp {
                (
                    _,
                    name,
                    CorpMembership {
                        allow_combat: combat,
                        priority: p,
                        ..
                    },
                ) => {
                    msg.push_str(&format!(
                        "| {:20} | {:7} | {:5} |\n",
                        name,
                        if *combat { "Y" } else { "N" },
                        p
                    ));
                }
            }
        }
    }

    ctx.trans.queue_for_session(ctx.session, Some(&msg)).await?;
    Ok(())
}

async fn corp_leave(ctx: &mut VerbContext<'_>, remaining: &str) -> UResult<()> {
    let user = get_user_or_fail(ctx)?;
    let (corpid, corp, mem) = match ctx
        .trans
        .match_user_corp_by_name(remaining, &user.username)
        .await?
    {
        None => user_error("Corp not found".to_owned())?,
        Some(v) => v,
    };
    if mem.joined_at.is_some() {
        ctx.trans
            .broadcast_to_corp(
                &corpid,
                &CorpCommType::Notice,
                None,
                &format!(
                    ansi!("There is a loud <red>boo<reset> as {} resigns from {}!\n"),
                    user.username, corp.name
                ),
            )
            .await?;
    } else {
        ctx.trans
            .queue_for_session(
                ctx.session,
                Some(&format!(
                    "You decline your invitation to join {}\n",
                    corp.name
                )),
            )
            .await?;
    }

    let mut delete_corp = false;
    if mem.permissions.contains(&CorpPermission::Holder) {
        let username_l = user.username.to_lowercase();
        let members = ctx.trans.list_corp_members(&corpid).await?;
        if members.len() == 1 {
            delete_corp = true;
        } else if !members.iter().any(|(name, mem)| {
            *name != username_l
                && mem.joined_at.is_some()
                && mem.permissions.contains(&CorpPermission::Holder)
        }) {
            user_error(
                "The last holder cannot resign from a non-empty \
                        corp - fire everyone else first, or promote a \
                        successor to holder"
                    .to_owned(),
            )?;
        }
    }
    ctx.trans
        .delete_corp_membership(&corpid, &user.username)
        .await?;
    if delete_corp {
        for dynzone in ctx
            .trans
            .find_dynzone_for_owner(&format!("corp/{}", &corp.name))
            .await?
        {
            recursively_destroy_or_move_item(&ctx.trans, &dynzone).await?;
        }
        ctx.trans.delete_corp(&corpid).await?;
    }
    Ok(())
}

async fn corp_fire(ctx: &mut VerbContext<'_>, remaining: &str) -> UResult<()> {
    let (target_raw, into_raw) = match remaining.rsplit_once(" from ") {
        None => user_error(
            ansi!("Usage: <bold>corp fire<reset> username <bold>from<reset> corpname").to_owned(),
        )?,
        Some(c) => c,
    };
    let user = get_user_or_fail(ctx)?;
    let player = get_player_item_or_fail(ctx).await?;
    let (corp_id, corp, mem) = match ctx
        .trans
        .match_user_corp_by_name(into_raw.trim(), &user.username)
        .await?
    {
        None => user_error("You don't seem to belong to a matching corp!".to_owned())?,
        Some(c) => c,
    };
    if !check_corp_perm(&CorpPermission::Fire, &mem) || mem.joined_at.is_none() {
        user_error("You don't have firing permissions for that corp".to_owned())?;
    }

    let target_user = search_item_for_user(
        ctx,
        &ItemSearchParams {
            include_all_players: true,
            ..ItemSearchParams::base(&player, target_raw.trim())
        },
    )
    .await?;

    if target_user.item_type != "player" {
        user_error("Only players can be fired.".to_owned())?;
    }

    if target_user.item_code == player.item_code {
        user_error("Fire yourself? You know you can just resign right?".to_owned())?;
    }

    match ctx
        .trans
        .match_user_corp_by_name(into_raw.trim(), &target_user.item_code)
        .await?
    {
        None => user_error(format!(
            "{} isn't currently hired.",
            &caps_first(&target_user.pronouns.subject)
        ))?,
        Some((
            _,
            _,
            CorpMembership {
                permissions: their_perm,
                joined_at: Some(their_join),
                ..
            },
        )) => {
            if their_perm.contains(&CorpPermission::Holder) {
                if !mem.permissions.contains(&CorpPermission::Holder) {
                    user_error(
                        "I love the ambition, but only holders can fire holders!".to_owned(),
                    )?;
                }
                if their_join < mem.joined_at.unwrap_or(Utc::now()) {
                    user_error(
                        "Whoah there young whippersnapper, holders can't fire more senior holders!"
                            .to_owned(),
                    )?;
                }
            }
        }
        _ => {}
    }

    ctx.trans
        .broadcast_to_corp(
            &corp_id,
            &CorpCommType::Notice,
            None,
            &format!(
                ansi!(
                    "A <blue>nervous silence<reset> falls across \
                        the workers as {} fires {} from {}.\n"
                ),
                user.username,
                target_user.display_for_sentence(1, false),
                corp.name
            ),
        )
        .await?;

    ctx.trans
        .delete_corp_membership(&corp_id, &target_user.item_code)
        .await?;
    Ok(())
}

async fn corp_promote(ctx: &mut VerbContext<'_>, remaining: &str) -> UResult<()> {
    let usage_error = || {
        user_error(
            ansi!("Usage: <bold>corp promote<reset> username <bold>in<reset> corpname <bold>to<reset> title <bold>privileges<reset> privileges\n\
                   Title can be any plain text up to 20 characters long.\n\
                   Permissions start with + or - (to add or take away) followed immediately by a permission name (e.g. holder, hire, fire, war)").to_owned()
    )
    };
    let (remaining, permissions_raw) = match remaining.rsplit_once(" privileges ") {
        None => usage_error()?,
        Some(c) => c,
    };
    let (target_raw, remaining) = match remaining.split_once(" in ") {
        None => usage_error()?,
        Some(c) => c,
    };
    let (corpname_raw, title_raw) = match remaining.split_once(" to ") {
        None => usage_error()?,
        Some(c) => c,
    };
    let title = ignore_special_characters(title_raw.trim());
    if title.len() > 20 {
        user_error("New title must be 20 characters or less".to_owned())?;
    }
    let psplit = permissions_raw.split(" ");
    let mut perm_add: BTreeSet<CorpPermission> = BTreeSet::new();
    let mut perm_rem: BTreeSet<CorpPermission> = BTreeSet::new();
    for perm in psplit {
        let add = if perm.starts_with("+") {
            true
        } else if perm.starts_with("-") {
            false
        } else {
            user_error(format!(
                "Expected {} to start with + or -",
                ignore_special_characters(perm)
            ))?
        };
        let perm = ignore_special_characters(&perm[1..]).to_lowercase();
        let perm = match CorpPermission::parse(&perm) {
            None => user_error(format!("Unknown permission {}", perm))?,
            Some(v) => v,
        };
        (if add { &mut perm_add } else { &mut perm_rem }).insert(perm);
    }
    match perm_add.intersection(&perm_rem).next() {
        Some(perm) => user_error(format!(
            "You tried to both add and remove privilege {} - make up your mind!",
            perm.display()
        ))?,
        None => {}
    }

    let user = get_user_or_fail(ctx)?;
    let player = get_player_item_or_fail(ctx).await?;
    let (corp_id, corp, mem) = match ctx
        .trans
        .match_user_corp_by_name(corpname_raw.trim(), &user.username)
        .await?
    {
        None => user_error("You don't seem to belong to a matching corp!".to_owned())?,
        Some(c) => c,
    };
    if !check_corp_perm(&CorpPermission::Promote, &mem) || mem.joined_at.is_none() {
        user_error("You don't have promote permissions for that corp".to_owned())?;
    }

    let target_user = search_item_for_user(
        ctx,
        &ItemSearchParams {
            include_all_players: true,
            ..ItemSearchParams::base(&player, target_raw.trim())
        },
    )
    .await?;

    if target_user.item_type != "player" {
        user_error("Only players can be promoted.".to_owned())?;
    }

    let mut their_mem = match ctx
        .trans
        .match_user_corp_by_name(corpname_raw.trim(), &target_user.item_code)
        .await?
    {
        None => user_error(format!(
            "{} isn't currently hired.",
            &caps_first(&target_user.pronouns.subject)
        ))?,
        Some((_, _, v)) => v,
    };
    match their_mem {
        CorpMembership {
            permissions: ref their_perm,
            joined_at: Some(their_join),
            ..
        } => {
            if their_perm.contains(&CorpPermission::Holder) {
                if !mem.permissions.contains(&CorpPermission::Holder) {
                    user_error(
                        "I love the ambition, but only holders can promote/demote holders!"
                            .to_owned(),
                    )?;
                }
                if their_join < mem.joined_at.unwrap_or(Utc::now()) {
                    user_error("Whoah there young whippersnapper, holders can't promote/demote more senior holders!".to_owned())?;
                }
            }
        }
        _ => {}
    }
    if target_user.item_code == player.item_code {
        if !mem.permissions.contains(&CorpPermission::Holder) {
            user_error("Only holders can promote / demote themselves".to_owned())?
        } else if perm_rem.contains(&CorpPermission::Holder) {
            let members = ctx.trans.list_corp_members(&corp_id).await?;
            if !members.iter().any(|(name, mem)| {
                *name != player.item_code
                    && mem.joined_at.is_some()
                    && mem.permissions.contains(&CorpPermission::Holder)
            }) {
                user_error(
                    "The last holder cannot demote themselves - \
                            promote a successor to holder first"
                        .to_owned(),
                )?
            }
        }
    }

    if !mem.permissions.contains(&CorpPermission::Holder)
        && !(&perm_add | &perm_rem).is_subset(&mem.permissions.clone().into_iter().collect())
    {
        user_error("You can only change permissions you have yourself.".to_owned())?
    }

    let perm_str_raw = perm_add
        .iter()
        .map(|v| "+".to_owned() + v.display())
        .join(" ")
        + " "
        + &perm_rem
            .iter()
            .map(|v| "-".to_owned() + v.display())
            .join(" ");

    ctx.trans.broadcast_to_corp(
        &corp_id,
        &CorpCommType::Notice, None,
        &format!("Everyone looks up from their desk as {} changes {}'s job title in {} to {} ({}).\n",
                 user.username,
                 target_user.display_for_sentence(1, false),
                 corp.name, &title, &perm_str_raw.trim())).await?;

    their_mem.job_title = title;
    their_mem.permissions = (&(&their_mem
        .permissions
        .clone()
        .into_iter()
        .collect::<BTreeSet<CorpPermission>>()
        | &perm_add)
        - &perm_rem)
        .into_iter()
        .collect();
    ctx.trans
        .upsert_corp_membership(&corp_id, &target_user.item_code, &their_mem)
        .await?;
    Ok(())
}

async fn corp_info(ctx: &mut VerbContext<'_>, remaining: &str) -> UResult<()> {
    let user = get_user_or_fail(ctx)?;
    let (corp_id, corp, _) = match ctx
        .trans
        .match_user_corp_by_name(remaining.trim(), &user.username)
        .await?
    {
        None => user_error("You don't seem to belong to a matching corp!".to_owned())?,
        Some(c) => c,
    };
    let mut msg = String::new();
    let founded_ago = humantime::format_duration(std::time::Duration::from_secs(
        (Utc::now() - corp.founded).num_seconds() as u64,
    ));
    msg.push_str(&format!(
        ansi!("<bold>{}'s essential information<reset>\nFounded: {} ago\nCredits: ${}\nTax rate: {}%\n"),
        &corp.name, &founded_ago, corp.credits, corp.tax as f64 * 1E-2
    ));
    let members = ctx.trans.list_corp_members(&corp_id).await?;
    if corp.allow_combat_required {
        msg.push_str(
            "Members ARE required to allow the corp to consent to fighting on their behalf",
        );
        let tot_mem = members.len();
        let mem_allow_combat = members.iter().filter(|(_, m)| m.allow_combat).count();
        if tot_mem == mem_allow_combat {
            msg.push_str(", and all members have done so.\n");
        } else {
            msg.push_str("- waiting on the following members to do so: ");
            for (u, _) in members.iter().filter(|(_, m)| !m.allow_combat) {
                msg.push_str(&caps_first(&u));
                msg.push(' ');
            }
            msg.push('\n');
        }
    } else {
        msg.push_str(
            "Members are NOT required to allow the corp to consent to fighting on their behalf.\n",
        );
    }

    msg.push_str(&format!(
        "Base member privileges: {}\n",
        &corp
            .member_permissions
            .iter()
            .map(|p| p.display())
            .join(" ")
    ));

    msg.push_str("Members:\n");
    msg.push_str(&format!(
        ansi!("<bgblue><white><bold>| {:20} | {:20} | {:20} |<reset>\n"),
        "Name", "Title", "Privileges"
    ));
    for (user, mem) in members {
        msg.push_str(&format!(
            ansi!("| {:20} | {:20} | {:20} |\n"),
            caps_first(&user),
            mem.job_title,
            mem.permissions.iter().map(|p| p.display()).join(" ")
        ));
    }

    ctx.trans
        .queue_for_session(&ctx.session, Some(&msg))
        .await?;
    Ok(())
}

async fn corp_allow_combat(ctx: &mut VerbContext<'_>, remaining: &str) -> UResult<()> {
    let (command, remaining) = parse_command_name(remaining);
    if command != "from" {
        user_error("Usage: corp allow combat from corpname".to_owned())?;
    }
    let user = get_user_or_fail(ctx)?;
    let (corp_id, corp, mut mem) = match ctx
        .trans
        .match_user_corp_by_name(remaining.trim(), &user.username)
        .await?
    {
        None => user_error("You don't seem to belong to a matching corp!".to_owned())?,
        Some(c) => c,
    };
    if mem.allow_combat {
        user_error(format!("You already allow combat for {}.", &corp.name))?;
    }
    mem.allow_combat = true;
    ctx.trans
        .upsert_corp_membership(&corp_id, &user.username.to_lowercase(), &mem)
        .await?;
    ctx.trans
        .queue_for_session(
            &ctx.session,
            Some("You now allow the corp to consent to fighting on your behalf.\n"),
        )
        .await?;
    Ok(())
}

async fn corp_disallow_combat(ctx: &mut VerbContext<'_>, remaining: &str) -> UResult<()> {
    let (command, remaining) = parse_command_name(remaining);
    if command != "from" {
        user_error("Usage: corp disallow combat from corpname".to_owned())?
    }
    let user = get_user_or_fail(ctx)?;
    let (corp_id, corp, mut mem) = match ctx
        .trans
        .match_user_corp_by_name(remaining.trim(), &user.username)
        .await?
    {
        None => user_error("You don't seem to belong to a matching corp!".to_owned())?,
        Some(c) => c,
    };
    if !mem.allow_combat {
        user_error("You already disallow combat for that corp.".to_owned())?;
    }
    if corp.allow_combat_required {
        user_error(format!(
            ansi!(
                "That corp requires everyone to allow combat as a condition \
                                  of employment, so you can't do that. You could do \
                                  <bold>corp resign {}<reset> instead."
            ),
            &corp.name
        ))?;
    }
    mem.allow_combat = false;
    ctx.trans
        .upsert_corp_membership(&corp_id, &user.username.to_lowercase(), &mem)
        .await?;
    ctx.trans
        .queue_for_session(
            &ctx.session,
            Some("You no longer allow the corp to consent to fighting on your behalf.\n"),
        )
        .await?;
    Ok(())
}

async fn corp_config(ctx: &mut VerbContext<'_>, remaining: &str) -> UResult<()> {
    let (corp_name_raw, remaining) = match remaining.split_once(" ") {
        None => user_error("Usage: corp config corpname ...".to_owned())?,
        Some(v) => v,
    };
    let remaining = remaining.trim().to_lowercase();
    let user = get_user_or_fail(ctx)?;

    let (corp_id, mut corp, mem) = match ctx
        .trans
        .match_user_corp_by_name(corp_name_raw.trim(), &user.username)
        .await?
    {
        None => user_error("You don't seem to belong to a matching corp!".to_owned())?,
        Some(c) => c,
    };
    if !check_corp_perm(&CorpPermission::Configure, &mem) {
        user_error("You don't have permission to change settings for that corp.".to_owned())?;
    }

    if remaining == "allow combat required" {
        corp.allow_combat_required = true;
        ctx.trans
            .broadcast_to_corp(
                &corp_id,
                &CorpCommType::Notice,
                None,
                &format!(
                    "{} announces a policy decision for {}: all new members must allow the \
                      corp to consent to fights on their behalf.\n",
                    &caps_first(&user.username),
                    &corp.name
                ),
            )
            .await?;
        ctx.trans.update_corp_details(&corp_id, &corp).await?;
    } else if remaining == "allow combat not required" {
        corp.allow_combat_required = false;
        ctx.trans.update_corp_details(&corp_id, &corp).await?;
        ctx.trans.broadcast_to_corp(
            &corp_id,
            &CorpCommType::Notice, None,
            &format!("{} announces a policy decision for {}: new members no longer need to allow the \
                      corp to consent to fights on their behalf. Any member can now use \
                      \"corp disallow combat from {}\" if they want to.\n",
                     &caps_first(&user.username),
                     &corp.name, &corp.name)).await?;
    } else if remaining.starts_with("base privileges ") {
        let remaining = remaining[("base privileges ".len())..].trim();
        let psplit = remaining.split(" ");
        let mut perms: BTreeSet<CorpPermission> = BTreeSet::new();
        for perm in psplit {
            let perm = ignore_special_characters(&perm.trim()).to_lowercase();
            if perm == "none" {
                continue;
            }
            let perm = match CorpPermission::parse(&perm) {
                None => user_error(format!("Unknown permission {}", perm))?,
                Some(p) if p == CorpPermission::Holder => {
                    user_error("You can't set holder as a base privilege.".to_owned())?
                }
                Some(p) if !check_corp_perm(&p, &mem) => {
                    user_error("You can't set base privilege you don't have yourself.".to_owned())?
                }
                Some(p) => p,
            };
            perms.insert(perm);
        }
        corp.member_permissions = perms.clone().into_iter().collect();
        let perm_str = &corp
            .member_permissions
            .iter()
            .map(|p| p.display())
            .join(" ");
        ctx.trans
            .broadcast_to_corp(
                &corp_id,
                &CorpCommType::Notice,
                None,
                &format!(
                    "{} announces a policy decision for {}: new members get \
                      the following privileges: {}.\n",
                    &caps_first(&user.username),
                    &corp.name,
                    if perm_str == "" { "none" } else { &perm_str }
                ),
            )
            .await?;
        ctx.trans.update_corp_details(&corp_id, &corp).await?;
    } else if remaining.starts_with("tax rate ") {
        let remaining = remaining[("tax rate ".len())..].trim();
        let remaining = match remaining.split_once("%") {
            None => remaining,
            Some((r, e)) => {
                if e != "" {
                    user_error("Tax rate must be a number".to_owned())?
                } else {
                    r.trim()
                }
            }
        };
        let rate: f32 = match remaining.parse::<f32>() {
            Err(_) => user_error("Tax rate must be a number".to_owned())?,
            Ok(r) => r,
        };
        if rate < 0.0 || rate >= 80.0 {
            user_error("Tax rate must be 0-80%".to_owned())?;
        }
        corp.tax = (rate * 100.0) as u16;
        ctx.trans.update_corp_details(&corp_id, &corp).await?;

        ctx.trans
            .broadcast_to_corp(
                &corp_id,
                &CorpCommType::Notice,
                None,
                &format!(
                    "{} has changed the tax rate for {} to {}%\n",
                    user.username,
                    corp.name,
                    (corp.tax as f64) / 100.0
                ),
            )
            .await?;
    } else {
        user_error("Configurations you can set:\n\tallow combat required\n\tallow combat not required\nbase permissions permission_name permission_name\ntax rate 12.34%".to_owned())?;
    }

    Ok(())
}

async fn corp_subscribe(ctx: &mut VerbContext<'_>, remaining: &str) -> UResult<()> {
    let user = get_user_or_fail(ctx)?;
    let (subs, remaining) = match remaining.split_once(" from ") {
        None => {
            let (_, corp, mem) = match ctx
                .trans
                .match_user_corp_by_name(remaining.trim(), &user.username)
                .await?
            {
                None => user_error("You don't seem to belong to a matching corp!".to_owned())?,
                Some(c) => c,
            };
            ctx.trans
                .queue_for_session(
                    &ctx.session,
                    Some(&format!(
                        "Subscriptions for {}: {}\n",
                        &corp.name,
                        &mem.comms_on.iter().map(|m| m.display()).join(" ")
                    )),
                )
                .await?;
            return Ok(());
        }
        Some(v) => v,
    };
    let mut subs_add = BTreeSet::<CorpCommType>::new();
    for sub in subs.trim().split(" ") {
        let sub = ignore_special_characters(&sub.trim());
        match CorpCommType::parse(&sub) {
            None => user_error(format!("Invalid commtype: {}", sub))?,
            Some(t) => subs_add.insert(t),
        };
    }
    let (corp_id, _, mut mem) = match ctx
        .trans
        .match_user_corp_by_name(remaining.trim(), &user.username)
        .await?
    {
        None => user_error("You don't seem to belong to a matching corp!".to_owned())?,
        Some(c) => c,
    };
    mem.comms_on = (&mem.comms_on.into_iter().collect::<BTreeSet<CorpCommType>>() | &subs_add)
        .into_iter()
        .collect();
    ctx.trans
        .upsert_corp_membership(&corp_id, &user.username, &mem)
        .await?;
    ctx.trans
        .queue_for_session(&ctx.session, Some("Subscriptions updated.\n"))
        .await?;
    Ok(())
}

async fn corp_unsubscribe(ctx: &mut VerbContext<'_>, remaining: &str) -> UResult<()> {
    let (subs, remaining) = match remaining.split_once(" from ") {
        None => user_error(
            "Usage: corp unsubscribe commtype commtype ... from corpname\n\
                            commtypes: chat notice connect reward death consent"
                .to_owned(),
        )?,
        Some(v) => v,
    };
    let mut subs_del = BTreeSet::<CorpCommType>::new();
    for sub in subs.trim().split(" ") {
        let sub = ignore_special_characters(&sub.trim());
        match CorpCommType::parse(&sub) {
            None => user_error(format!("Invalid commtype: {}", sub))?,
            Some(t) => subs_del.insert(t),
        };
    }
    let user = get_user_or_fail(ctx)?;
    let (corp_id, _, mut mem) = match ctx
        .trans
        .match_user_corp_by_name(remaining.trim(), &user.username)
        .await?
    {
        None => user_error("You don't seem to belong to a matching corp!".to_owned())?,
        Some(c) => c,
    };
    mem.comms_on = (&mem.comms_on.into_iter().collect::<BTreeSet<CorpCommType>>() - &subs_del)
        .into_iter()
        .collect();
    ctx.trans
        .upsert_corp_membership(&corp_id, &user.username, &mem)
        .await?;
    ctx.trans
        .queue_for_session(&ctx.session, Some("Subscriptions updated.\n"))
        .await?;
    Ok(())
}

async fn corp_order(ctx: &mut VerbContext<'_>, remaining: &str) -> UResult<()> {
    let (corpname, number_s) = match remaining.split_once(" as ") {
        None => user_error("Usage: corp order corpname as number".to_owned())?,
        Some(v) => v,
    };
    let number = match nom::character::complete::i64::<&str, ()>(number_s) {
        Ok((rest, num)) if rest.trim() == "" => num,
        _ => user_error("Usage: corp order corpname as number".to_owned())?,
    };
    let user = get_user_or_fail(ctx)?;
    let (corp_id, _, mut mem) = match ctx
        .trans
        .match_user_corp_by_name(corpname.trim(), &user.username)
        .await?
    {
        None => user_error("You don't seem to belong to a matching corp!".to_owned())?,
        Some(c) => c,
    };
    mem.priority = number;
    ctx.trans
        .upsert_corp_membership(&corp_id, &user.username, &mem)
        .await?;
    ctx.trans
        .queue_for_session(&ctx.session, Some("Updated!\n"))
        .await?;
    Ok(())
}

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(
        self: &Self,
        ctx: &mut VerbContext,
        _verb: &str,
        remaining: &str,
    ) -> UResult<()> {
        let (command, remaining) = parse_command_name(remaining);
        match command.to_lowercase().as_str() {
            "hire" | "invite" => corp_invite(ctx, remaining).await?,
            "join" => corp_join(ctx, remaining).await?,
            "" | "list" => corp_list(ctx, remaining).await?,
            "leave" | "resign" => corp_leave(ctx, remaining).await?,
            "fire" | "dismiss" => corp_fire(ctx, remaining).await?,
            "promote" | "demote" => corp_promote(ctx, remaining).await?,
            "info" => corp_info(ctx, remaining).await?,
            "allow" => {
                let (command, remaining) = parse_command_name(remaining);
                if command.to_lowercase() != "combat" {
                    user_error("Allow must be followed with combat".to_owned())?
                } else {
                    corp_allow_combat(ctx, remaining).await?
                }
            }
            "disallow" => {
                let (command, remaining) = parse_command_name(remaining);
                if command.to_lowercase() != "combat" {
                    user_error("Disallow must be followed with combat".to_owned())?
                } else {
                    corp_disallow_combat(ctx, remaining).await?
                }
            }
            "configure" | "config" => {
                corp_config(ctx, remaining).await?;
            }
            "sub" | "subscribe" => {
                corp_subscribe(ctx, remaining).await?;
            }
            "unsub" | "unsubscribe" => {
                corp_unsubscribe(ctx, remaining).await?;
            }
            "order" => {
                corp_order(ctx, remaining).await?;
            }
            _ => user_error("Unknown command".to_owned())?,
        }
        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
