use super::{
    get_player_item_or_fail, parsing::parse_count, search_items_for_user, user_error,
    ItemSearchParams, UResult, UserError, UserVerb, UserVerbRef, VerbContext,
};
use crate::{
    models::item::LocationActionType,
    regular_tasks::queued_command::{
        queue_command, QueueCommand, QueueCommandHandler, QueuedCommandContext,
    },
    services::{
        comms::broadcast_to_room,
        urges::{hunger_changed, thirst_changed},
    },
    static_content::possession_type::possession_data,
};
use ansi::ansi;
use async_trait::async_trait;
use std::time;

pub struct QueueHandler;
#[async_trait]
impl QueueCommandHandler for QueueHandler {
    async fn start_command(&self, ctx: &mut QueuedCommandContext<'_>) -> UResult<time::Duration> {
        if ctx.item.death_data.is_some() {
            user_error(
                "You try to eat it, but your ghostly hands slip through it uselessly".to_owned(),
            )?;
        }
        let item_id = match ctx.command {
            QueueCommand::Eat { possession_id } => possession_id,
            _ => user_error("Unexpected command".to_owned())?,
        };
        let item = match ctx
            .trans
            .find_item_by_type_code("possession", &item_id)
            .await?
        {
            None => user_error("Item not found".to_owned())?,
            Some(it) => it,
        };
        if item.location != format!("{}/{}", &ctx.item.item_type, &ctx.item.item_code) {
            user_error(format!(
                "You try to eat {} but realise you no longer have it",
                item.display_for_sentence(1, false)
            ))?
        }
        let msg = format!(
            "{} prepares to eat {}\n",
            &ctx.item.display_for_sentence(1, true),
            &item.display_for_sentence(1, false)
        );
        broadcast_to_room(ctx.trans, &ctx.item.location, None, &msg).await?;
        Ok(time::Duration::from_secs(1))
    }

    #[allow(unreachable_patterns)]
    async fn finish_command(&self, ctx: &mut QueuedCommandContext<'_>) -> UResult<()> {
        if ctx.item.death_data.is_some() {
            user_error(
                "You try to eat it, but your ghostly hands slip through it uselessly".to_owned(),
            )?;
        }
        let item_id = match ctx.command {
            QueueCommand::Eat { possession_id } => possession_id,
            _ => user_error("Unexpected command".to_owned())?,
        };

        let item = match ctx
            .trans
            .find_item_by_type_code("possession", &item_id)
            .await?
        {
            None => user_error("Item not found".to_owned())?,
            Some(it) => it,
        };
        if item.location != format!("{}/{}", &ctx.item.item_type, &ctx.item.item_code) {
            user_error(format!(
                "You try to eat {} but realise you no longer have it!",
                &item.display_for_sentence(1, false)
            ))?
        }
        if item.action_type == LocationActionType::Worn {
            user_error(
                ansi!("You're wearing it - try using <bold>remove<reset> first").to_owned(),
            )?;
        }

        let possession_data = match item
            .possession_type
            .as_ref()
            .and_then(|pt| possession_data().get(&pt))
        {
            None => {
                user_error("That item no longer exists in the game so can't be handled".to_owned())?
            }
            Some(pd) => pd,
        };

        let eat_data = possession_data
            .eat_data
            .as_ref()
            .ok_or_else(|| UserError("You can't eat that!".to_owned()))?;

        let urges = ctx
            .item
            .urges
            .as_ref()
            .ok_or_else(|| UserError("You don't seem to have the appetite.".to_owned()))?;
        if (urges.hunger.value as i16) < -eat_data.hunger_impact {
            user_error("You don't seem to have the appetite.".to_owned())?;
        }
        let how_many_left = (if item.charges <= 1 { 1 } else { item.charges }) as u16;
        let how_many_to_fill = if eat_data.hunger_impact >= 0 {
            1
        } else {
            urges.hunger.value / ((-eat_data.hunger_impact) as u16)
        };
        let how_many_eaten = how_many_to_fill.min(how_many_left).max(1);

        let msg = format!(
            "{} {} {}\n",
            &ctx.item.display_for_sentence(1, true),
            &(if how_many_eaten == how_many_left {
                "polishes off".to_owned()
            } else {
                format!(
                    "eats {} bite{} from",
                    how_many_eaten,
                    if how_many_eaten == 1 { "" } else { "s" }
                )
            }),
            &item.display_for_sentence(1, false)
        );
        broadcast_to_room(ctx.trans, &ctx.item.location, None, &msg).await?;

        if let Some(urges) = ctx.item.urges.as_mut() {
            urges.hunger.last_value = urges.hunger.value;
            urges.hunger.value = (urges.hunger.value as i16
                + (how_many_eaten as i16) * eat_data.hunger_impact)
                .clamp(0, 10000) as u16;
            urges.thirst.last_value = urges.thirst.value;
            urges.thirst.value = (urges.thirst.value as i16
                + (how_many_eaten as i16) * eat_data.thirst_impact)
                .clamp(0, 10000) as u16;
        }
        hunger_changed(&ctx.trans, &ctx.item).await?;
        thirst_changed(&ctx.trans, &ctx.item).await?;

        if item.charges <= (how_many_eaten as u8) {
            ctx.trans.delete_item("possession", &item_id).await?;
        } else {
            let mut item_mut = (*item).clone();
            item_mut.charges -= how_many_eaten as u8;
            ctx.trans.save_item_model(&item_mut).await?;
        }
        Ok(())
    }
}

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(
        self: &Self,
        ctx: &mut VerbContext,
        _verb: &str,
        mut remaining: &str,
    ) -> UResult<()> {
        let player_item = get_player_item_or_fail(ctx).await?;

        let mut eat_limit = Some(1);
        if remaining == "all" || remaining.starts_with("all ") {
            remaining = remaining[3..].trim();
            eat_limit = None;
        } else if let (Some(n), remaining2) = parse_count(remaining) {
            eat_limit = Some(n);
            remaining = remaining2;
        }
        let targets = search_items_for_user(
            ctx,
            &ItemSearchParams {
                include_contents: true,
                item_type_only: Some("possession"),
                limit: eat_limit.unwrap_or(100),
                ..ItemSearchParams::base(&player_item, &remaining)
            },
        )
        .await?;

        if player_item.death_data.is_some() {
            user_error(
                "You try to eat it, but your ghostly hands slip through it uselessly".to_owned(),
            )?;
        }

        let mut player_item_mut = (*player_item).clone();
        for target in targets {
            if target.item_type != "possession" {
                user_error("You can't eat that!".to_owned())?;
            }
            target
                .possession_type
                .as_ref()
                .and_then(|pt| possession_data().get(pt))
                .and_then(|pd| pd.eat_data.as_ref())
                .ok_or_else(|| UserError("You can't eat that!".to_owned()))?;
            queue_command(
                ctx,
                &mut player_item_mut,
                &QueueCommand::Eat {
                    possession_id: target.item_code.clone(),
                },
            )
            .await?;
        }
        ctx.trans.save_item_model(&player_item_mut).await?;
        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
