use std::collections::BTreeMap;

use super::{CommandHandlingError::UserError, UResult, UserVerb, UserVerbRef, VerbContext};
use ansi_markup::parse_ansi_markup;
use async_trait::async_trait;
use once_cell::sync::OnceCell;
use serde_yaml::from_str as from_yaml_str;

fn load_help_yaml(input: &str) -> BTreeMap<String, String> {
    let mut map: BTreeMap<String, String> = from_yaml_str(input).unwrap();
    for val in map.values_mut() {
        *val = parse_ansi_markup(val).unwrap();
    }
    map
}

fn always_help_pages() -> &'static BTreeMap<String, String> {
    static CELL: OnceCell<BTreeMap<String, String>> = OnceCell::new();

    CELL.get_or_init(|| load_help_yaml(include_str!("help/always.yaml")))
}

fn registered_help_pages() -> &'static BTreeMap<String, String> {
    static CELL: OnceCell<BTreeMap<String, String>> = OnceCell::new();

    CELL.get_or_init(|| load_help_yaml(include_str!("help/registered.yaml")))
}

fn unregistered_help_pages() -> &'static BTreeMap<String, String> {
    static CELL: OnceCell<BTreeMap<String, String>> = OnceCell::new();

    CELL.get_or_init(|| load_help_yaml(include_str!("help/unregistered.yaml")))
}

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(
        self: &Self,
        ctx: &mut VerbContext,
        _verb: &str,
        remaining: &str,
    ) -> UResult<()> {
        let mut help = None;
        let is_unregistered = match ctx.user_dat {
            None => true,
            Some(user_dat) => !user_dat.terms.terms_complete,
        };
        let remaining = remaining.trim();
        if is_unregistered {
            help = help.or_else(|| unregistered_help_pages().get(remaining));
        } else {
            help = help.or_else(|| registered_help_pages().get(remaining));
        }
        help = help.or_else(|| always_help_pages().get(remaining).map(|v| v));
        let help_final = help.ok_or(UserError("No help available on that".to_string()))?;
        ctx.trans
            .queue_for_session(ctx.session, Some(&(help_final.clone() + "\n")))
            .await?;
        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn registered_help_ok() {
        registered_help_pages();
    }

    #[test]
    fn unregistered_help_ok() {
        unregistered_help_pages();
    }

    #[test]
    fn always_help_ok() {
        always_help_pages();
    }
}
