use super::{
    drop::consider_expire_job_for_item, get_player_item_or_fail, user_error, UResult, UserVerb,
    UserVerbRef, VerbContext,
};
use crate::{
    models::item::{LocationActionType, Scavtype, SkillType},
    regular_tasks::queued_command::{
        queue_command_and_save, QueueCommand, QueueCommandHandler, QueuedCommandContext,
    },
    services::{
        capacity::{check_item_capacity, CapacityLevel},
        comms::broadcast_to_room,
        skills::skill_check_and_grind,
    },
};
use ansi::ansi;
use async_trait::async_trait;
use std::time;

pub struct QueueHandler;
#[async_trait]
impl QueueCommandHandler for QueueHandler {
    async fn start_command(&self, ctx: &mut QueuedCommandContext<'_>) -> UResult<time::Duration> {
        if ctx.item.death_data.is_some() {
            user_error(
                "You try to search the area, but you realise your ghost hands aren't good for searching".to_owned(),
            )?;
        }
        broadcast_to_room(
            &ctx.trans,
            &ctx.item.location,
            None,
            &format!(
                ansi!("<blue>{} starts methodically searching the area.<reset>\n"),
                ctx.item.display_for_sentence(1, true),
            ),
        )
        .await?;
        Ok(time::Duration::from_secs(3))
    }

    #[allow(unreachable_patterns)]
    async fn finish_command(&self, ctx: &mut QueuedCommandContext<'_>) -> UResult<()> {
        if ctx.item.death_data.is_some() {
            user_error(
                "You try to search the area, but you realise your ghost hands aren't good for searching".to_owned(),
            )?;
        }
        let scav = ctx
            .item
            .total_skills
            .get(&SkillType::Scavenge)
            .unwrap_or(&0.0)
            .clone();
        let found_opt = ctx
            .trans
            .find_scavhidden_item_by_location(
                &ctx.item.location,
                &Scavtype::Scavenge,
                (scav * 100.0).max(0.0) as i64,
            )
            .await?;
        let mut found = match found_opt {
            None => user_error("You have a look, and you're pretty sure there's nothing to find here. [Try searching elsewhere, or come back later]".to_owned())?,
            Some(v) => v,
        };
        let diff: f64 = (match found.action_type {
            LocationActionType::Scavhidden { difficulty, .. } => difficulty,
            _ => 800,
        }) as f64
            / 100.0;

        if skill_check_and_grind(&ctx.trans, &mut ctx.item, &SkillType::Scavenge, diff).await? < 0.0
        {
            // No crit fail for now, it is either yes or no.
            match ctx.get_session().await? {
                None => {},
                Some((sess, _)) =>
                    ctx.trans.queue_for_session(&sess, Some("You give up searching, but you still have a feeling there is something here.\n")).await?
            }
            return Ok(());
        }

        found.action_type = LocationActionType::Normal;
        match check_item_capacity(&ctx.trans, &ctx.item, found.weight).await? {
            CapacityLevel::OverBurdened | CapacityLevel::AboveItemLimit => {
                consider_expire_job_for_item(&ctx.trans, &found).await?;
                broadcast_to_room(
                    &ctx.trans,
                    &ctx.item.location,
                    None,
                    &format!("{} seems to have found {} - but can't carry it so drops it on the ground.\n",
                            ctx.item.display_for_sentence(1, true),
                            found.display_for_sentence(1, false),
                    ),
                ).await?;
            }
            _ => {
                broadcast_to_room(
                    &ctx.trans,
                    &ctx.item.location,
                    None,
                    &format!(
                        "{} seems to have found {}.\n",
                        ctx.item.display_for_sentence(1, true),
                        found.display_for_sentence(1, false),
                    ),
                )
                .await?;
                found.location = ctx.item.refstr();
            }
        }
        ctx.trans.save_item_model(&found).await?;

        Ok(())
    }
}

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(
        self: &Self,
        ctx: &mut VerbContext,
        _verb: &str,
        _remaining: &str,
    ) -> UResult<()> {
        let player_item = get_player_item_or_fail(ctx).await?;
        queue_command_and_save(ctx, &player_item, &QueueCommand::Scavenge).await?;
        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
