use super::{get_player_item_or_fail, user_error, UResult, UserVerb, UserVerbRef, VerbContext};
use crate::{
    language,
    models::{
        item::{SkillType, StatType},
        user::{wristpad_hack_data, xp_to_hack_slots},
    },
};
use ansi::ansi;
use async_trait::async_trait;

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(
        self: &Self,
        ctx: &mut VerbContext,
        _verb: &str,
        _remaining: &str,
    ) -> UResult<()> {
        let player_item = get_player_item_or_fail(ctx).await?;

        let user = match ctx.user_dat {
            None => user_error("Log in first".to_owned())?,
            Some(user) => user,
        };

        let mut msg = String::new();
        msg.push_str(&format!(
            ansi!("<bgblue><white><bold>| {:11} | {:5} | {:5} |<reset>\n"),
            "Stat", "Raw", "Total"
        ));
        for st in StatType::values().iter() {
            msg.push_str(&format!(
                ansi!("| <bold>{:11}<reset> | {:5.2} | {:5.2} |\n"),
                st.display(),
                user.raw_stats.get(st).unwrap_or(&0.0),
                player_item.total_stats.get(st).unwrap_or(&0.0)
            ));
        }
        msg.push_str("\n");

        msg.push_str(&format!(
            ansi!("<bgblue><white><bold>| {:11} | {:5} | {:5} |<reset>\n"),
            "Skill", "Raw", "Total"
        ));
        for st in SkillType::values().iter() {
            msg.push_str(&format!(
                ansi!("| <bold>{:11}<reset> | {:5.2} | {:5.2} |\n"),
                st.display(),
                user.raw_skills.get(st).unwrap_or(&0.0),
                player_item.total_skills.get(st).unwrap_or(&0.0)
            ));
        }
        msg.push_str("\n");
        msg.push_str(&format!(
            ansi!(
                "Experience: <bold>{}<reset> total, \
                                     change of {} this re-roll, \
                                     {} spent since re-roll.\n"
            ),
            player_item.total_xp,
            user.experience.xp_change_for_this_reroll,
            user.experience.spent_xp
        ));

        if !user.wristpad_hacks.is_empty() {
            let hack_names = user
                .wristpad_hacks
                .iter()
                .map(|h| {
                    wristpad_hack_data()
                        .get(h)
                        .map(|hd| hd.name)
                        .unwrap_or("UNKNOWN")
                })
                .collect::<Vec<&'static str>>();
            msg.push_str(&format!(
                "You have hacks installed on your wristpad: {}\n",
                &language::join_words(&hack_names)
            ));
        }

        let hack_slots = xp_to_hack_slots(player_item.total_xp) as usize;
        if hack_slots > user.wristpad_hacks.len() {
            let free_slots = hack_slots - user.wristpad_hacks.len();
            msg.push_str(&format!(
                "You have {} free hack slot{} on your wristpad.\n",
                free_slots,
                if free_slots == 1 { "" } else { "s" }
            ));
        } else {
            msg.push_str("You have no free hack slots on your wristpad.\n");
        };

        ctx.trans.queue_for_session(ctx.session, Some(&msg)).await?;

        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
