use crate::{
    db::ItemSearchParams,
    models::item::{ConversationIntensity, ConversationalStyle},
    services::sharing::{
        change_conversation_intensity, change_conversation_topic, change_conversational_style,
        display_conversation_status, parse_conversation_topic, start_conversation,
    },
};

use super::{
    get_player_item_or_fail, search_item_for_user, user_error, UResult, UserError, UserVerb,
    UserVerbRef, VerbContext,
};
use ansi::ansi;
use async_trait::async_trait;

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(
        self: &Self,
        ctx: &mut VerbContext,
        verb: &str,
        remaining: &str,
    ) -> UResult<()> {
        let mut player_item = (*(get_player_item_or_fail(ctx).await?)).clone();
        if player_item.death_data.is_some() {
            user_error("You can't do that, you're dead!".to_string())?;
        }

        let remaining = remaining.trim();
        if let Some(ac) = player_item.active_conversation.as_ref() {
            match ConversationalStyle::from_name(verb) {
                None => {}
                Some(style) => {
                    change_conversational_style(ctx, &mut player_item, style).await?;
                    ctx.trans.save_item_model(&player_item).await?;
                    return Ok(());
                }
            }
            match parse_conversation_topic(&format!("{} {}", verb, remaining))
                .map_err(|e| UserError(e.to_owned()))?
            {
                None => {}
                Some(topic) => {
                    change_conversation_topic(ctx, &mut player_item, topic).await?;
                    ctx.trans.save_item_model(&player_item).await?;
                    return Ok(());
                }
            }

            if verb == "share" {
                match ConversationIntensity::from_adverb(remaining) {
                    None => {}
                    Some(intensity) => {
                        change_conversation_intensity(ctx, &mut player_item, intensity).await?;
                        ctx.trans.save_item_model(&player_item).await?;
                        return Ok(());
                    }
                }
                if remaining == "status" {
                    let (partner_type, partner_code) = ac
                        .partner_ref
                        .split_once("/")
                        .ok_or_else(|| UserError("Bad share partner".to_owned()))?;
                    if let Some(partner) = ctx
                        .trans
                        .find_item_by_type_code(partner_type, partner_code)
                        .await?
                    {
                        display_conversation_status(&ctx.trans, &player_item, &partner).await?;
                        return Ok(());
                    }
                }
            }

            user_error("You're already sharing knowledge!".to_owned())?;
        }

        let (word2, remaining) = remaining.split_once(" ").ok_or_else(|| {
            UserError(
                ansi!("Start your encounter with the <bold>share knowledge with<reset> command first.")
                    .to_owned(),
            )
        })?;
        let (word3, remaining) = remaining.trim().split_once(" ").ok_or_else(|| {
            UserError(ansi!("<bold>share knowledge with<reset> command first.").to_owned())
        })?;
        if verb != "share" || word2.trim() != "knowledge" || word3.trim() != "with" {
            user_error(
                ansi!("Start your encounter with the <bold>share knowledge with<reset> command first.")
                    .to_owned(),
            )?;
        }

        let with_whom = search_item_for_user(
            ctx,
            &ItemSearchParams {
                include_loc_contents: true,
                limit: 1,
                ..ItemSearchParams::base(&player_item, remaining)
            },
        )
        .await?;

        start_conversation(&ctx.trans, &player_item, &with_whom).await?;

        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
