use super::{
    corp::check_corp_perm, get_player_item_or_fail, user_error, UResult, UserError, UserVerb,
    UserVerbRef, VerbContext,
};
use crate::{
    models::{
        corp::{CorpCommType, CorpPermission},
        item::{Item, ItemSpecialData},
    },
    static_content::room::{room_map_by_code, Direction, RentSuiteType},
};
use ansi::ansi;
use async_trait::async_trait;
use chrono::{Duration, Utc};
use itertools::Itertools;

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(
        self: &Self,
        ctx: &mut VerbContext,
        _verb: &str,
        remaining: &str,
    ) -> UResult<()> {
        let remaining = remaining.trim();
        let (item_name, corp_name) = match remaining.split_once(" for ") {
            None => (remaining, None),
            Some((i, c)) => (i.trim(), Some(c.trim())),
        };
        let player_item = get_player_item_or_fail(ctx).await?;
        let (loc_type, loc_code) = player_item
            .location
            .split_once("/")
            .ok_or_else(|| UserError("Invalid location".to_owned()))?;
        if loc_type != "room" {
            user_error(
                "You must go to where you rented the place (e.g. reception) to vacate.".to_owned(),
            )?;
        }
        let room = room_map_by_code()
            .get(loc_code)
            .ok_or_else(|| UserError("Can't find your room".to_owned()))?;
        if room.rentable_dynzone.is_empty() {
            user_error("Go to where you rented the place (e.g. reception) to vacate.".to_owned())?;
        }
        let rentinfo = match room
            .rentable_dynzone
            .iter()
            .find(|ri| ri.rent_what == item_name)
        {
            None => user_error(format!(
                "Vacate must be followed by the specific thing you want to vacate: {}",
                room.rentable_dynzone
                    .iter()
                    .map(|ri| ri.rent_what.as_str())
                    .join(", ")
            ))?,
            Some(v) => v,
        };
        let corp = match (&rentinfo.suite_type, corp_name) {
            (RentSuiteType::Commercial, None) =>
                user_error(format!(
                    ansi!("This is a commercial suite, you need to vacate it using the name of the corp. Try <bold>vacate {} for corpname<reset>"),
                    item_name
                ))?,
            (RentSuiteType::Residential, Some(_)) =>
                user_error("This is a residential suite, you can't vacate it for a corp. Try <bold>vacate {}<reset>".to_owned())?,
            (RentSuiteType::Residential, None) => None,
            (RentSuiteType::Commercial, Some(n)) => match ctx.trans.match_user_corp_by_name(&n, &player_item.item_code).await? {
                None => user_error("I can't find that corp in your list of corps!".to_owned())?,
                Some((_, _, mem)) if !check_corp_perm(&CorpPermission::Holder, &mem) => user_error("You don't have holder permissions in that corp.".to_owned())?,
                Some((corp_id, corp, _)) => Some((corp_id, corp))
            },
        };

        let exit = Direction::IN {
            item: corp
                .as_ref()
                .map(|c| c.1.name.clone())
                .unwrap_or_else(|| player_item.display.clone()),
        };

        match ctx
            .trans
            .find_exact_dyn_exit(&player_item.location, &exit)
            .await?
            .as_ref()
            .and_then(|it| it.location.split_once("/"))
        {
            None => user_error("You aren't renting anything from here!".to_owned())?,
            Some((ref ex_zone_t, ref ex_zone_c)) => {
                if let Some(ex_zone) = ctx
                    .trans
                    .find_item_by_type_code(ex_zone_t, ex_zone_c)
                    .await?
                {
                    match ex_zone.special_data {
                        Some(ItemSpecialData::DynzoneData {
                            vacate_after: Some(_),
                            ..
                        }) => user_error("Your lease is already up for termination.".to_owned())?,
                        Some(ItemSpecialData::DynzoneData {
                            vacate_after: None,
                            zone_exit: ref ex,
                        }) => {
                            ctx.trans
                                .save_item_model(&Item {
                                    special_data: Some(ItemSpecialData::DynzoneData {
                                        zone_exit: ex.clone(),
                                        vacate_after: Some(Utc::now() + Duration::days(1)),
                                    }),
                                    ..(*ex_zone).clone()
                                })
                                .await?;
                            ctx.trans.queue_for_session(ctx.session, Some("The robot files away your notice of intention to vacate. \"You have 24 hours to get all your stuff out, then the landlord will send someone up to boot out anyone still in there, and we will sell anything left behind to cover our costs. If you change your mind before then, just rent again and we'll cancel out your notice and let you keep the same apartment - then you'll have to pay the setup fee again though.\"\n")).await?;
                            match corp {
                                None => {},
                                Some(corptup) =>
                                    ctx.trans.broadcast_to_corp(
                                        &corptup.0,
                                        &CorpCommType::Notice,
                                        Some(&player_item.item_code),
                                        &format!(
                                            ansi!("<cyan>[{}] {} just gave notice to vacate {}! The landlord replied with: \"You have 24 hours to get all your stuff out, then the landlord will send someone up to boot out anyone still in there, and we will sell anything left behind to cover our costs. If you change your mind before then, just rent again and we'll cancel out your notice and let you keep the same apartment - then you'll have to pay the setup fee again though.\"<reset>\n"),
                                            &corptup.1.name,
                                            &player_item.display_for_sentence(1, false),
                                            &ex_zone.display
                                        )
                                    ).await?
                            }
                        }
                        _ => user_error("The premises seem to be broken anyway".to_owned())?,
                    }
                }
            }
        }

        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
