use super::{
    get_player_item_or_fail, search_item_for_user, user_error, ItemSearchParams, UResult, UserVerb,
    UserVerbRef, VerbContext,
};
use crate::{
    models::item::{LocationActionType, SkillType},
    regular_tasks::queued_command::{
        queue_command_and_save, QueueCommand, QueueCommandHandler, QueuedCommandContext,
    },
    services::{comms::broadcast_to_room, skills::skill_check_and_grind},
    static_content::possession_type::possession_data,
};
use async_trait::async_trait;
use std::time;

pub struct QueueHandler;
#[async_trait]
impl QueueCommandHandler for QueueHandler {
    async fn start_command(&self, ctx: &mut QueuedCommandContext<'_>) -> UResult<time::Duration> {
        if ctx.item.death_data.is_some() {
            user_error(
                "You try to wield it, but your ghostly hands slip through it uselessly".to_owned(),
            )?;
        }
        let item_id = match ctx.command {
            QueueCommand::Wield { possession_id } => possession_id,
            _ => user_error("Unexpected command".to_owned())?,
        };
        let item = match ctx
            .trans
            .find_item_by_type_code("possession", &item_id)
            .await?
        {
            None => user_error("Item not found".to_owned())?,
            Some(it) => it,
        };
        if item.location != format!("player/{}", ctx.item.item_code) {
            user_error("You try to wield it but realise you no longer have it".to_owned())?
        }
        let msg = format!(
            "{} fumbles around with {} {}\n",
            &ctx.item.display_for_sentence(1, true),
            &ctx.item.pronouns.possessive,
            &item.display_for_sentence(1, false)
        );
        broadcast_to_room(ctx.trans, &ctx.item.location, None, &msg).await?;
        let mut draw_level: f64 = ctx
            .item
            .total_skills
            .get(&SkillType::Quickdraw)
            .to_owned()
            .unwrap_or(&8.0)
            .clone();

        let skill_result =
            skill_check_and_grind(ctx.trans, ctx.item, &SkillType::Quickdraw, draw_level).await?;
        if skill_result < -0.5 {
            draw_level -= 2.0;
        } else if skill_result < -0.25 {
            draw_level -= 1.0;
        } else if skill_result > 0.5 {
            draw_level += 2.0;
        } else if skill_result > 0.25 {
            draw_level += 1.0;
        }

        let wait_ticks = (12.0 - (draw_level / 2.0)).min(8.0).max(1.0);
        Ok(time::Duration::from_millis(
            (wait_ticks * 500.0).round() as u64
        ))
    }

    #[allow(unreachable_patterns)]
    async fn finish_command(&self, ctx: &mut QueuedCommandContext<'_>) -> UResult<()> {
        if ctx.item.death_data.is_some() {
            user_error(
                "You try to wield it, but your ghostly hands slip through it uselessly".to_owned(),
            )?;
        }
        let item_id = match ctx.command {
            QueueCommand::Wield { possession_id } => possession_id,
            _ => user_error("Unexpected command".to_owned())?,
        };
        let item = match ctx
            .trans
            .find_item_by_type_code("possession", &item_id)
            .await?
        {
            None => user_error("Item not found".to_owned())?,
            Some(it) => it,
        };
        if item.location != format!("player/{}", ctx.item.item_code) {
            user_error("You try to wield it but realise you no longer have it".to_owned())?
        }
        let msg = format!(
            "{} wields {}\n",
            &ctx.item.display_for_sentence(1, true),
            &item.display_for_sentence(1, false)
        );
        broadcast_to_room(ctx.trans, &ctx.item.location, None, &msg).await?;
        ctx.trans
            .set_exclusive_action_type_to(
                &item,
                &LocationActionType::Wielded,
                &LocationActionType::Normal,
            )
            .await?;
        Ok(())
    }
}

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(
        self: &Self,
        ctx: &mut VerbContext,
        _verb: &str,
        remaining: &str,
    ) -> UResult<()> {
        let player_item = get_player_item_or_fail(ctx).await?;
        let weapon = search_item_for_user(
            ctx,
            &ItemSearchParams {
                include_contents: true,
                limit: 1,
                ..ItemSearchParams::base(&player_item, &remaining)
            },
        )
        .await?;
        if player_item.death_data.is_some() {
            user_error(
                "You try to wield it, but your ghostly hands slip through it uselessly".to_owned(),
            )?;
        }
        if weapon.action_type == LocationActionType::Wielded {
            user_error("You're actually already wielding it.".to_owned())?;
        }
        if weapon.item_type != "possession"
            || weapon
                .possession_type
                .as_ref()
                .and_then(|poss_type| possession_data().get(&poss_type))
                .and_then(|poss_data| poss_data.weapon_data.as_ref())
                .is_none()
        {
            user_error("You can't wield that!".to_owned())?;
        }
        queue_command_and_save(
            ctx,
            &player_item,
            &QueueCommand::Wield {
                possession_id: weapon.item_code.clone(),
            },
        )
        .await?;
        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
