use super::{TaskHandler, TaskRunContext};
#[double]
use crate::db::DBTrans;
use crate::message_handler::user_commands::{
    close, cut, drink, drop, eat, fill, get, improvise, make, movement, open, put, recline, remove,
    scavenge, sit, stand, use_cmd, user_error, wear, wield, CommandHandlingError, UResult,
    VerbContext,
};
use crate::message_handler::ListenerSession;
use crate::models::session::Session;
use crate::models::{
    item::Item,
    task::{Task, TaskDetails, TaskMeta},
};
use crate::static_content::{possession_type::PossessionType, room::Direction};
use crate::DResult;
use async_trait::async_trait;
use chrono::Utc;
use mockall_double::double;
use once_cell::sync::OnceCell;
use serde::{Deserialize, Serialize};
use std::collections::{BTreeMap, BTreeSet};
use std::time;
use uuid::Uuid;

#[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize, Ord, PartialOrd)]
pub enum MovementSource {
    Command {
        event_id: Uuid,
    },
    Follow {
        origin_uuid: Uuid,
        chain: Vec<String>,
    },
    Internal {
        event_id: Uuid,
    },
}

impl MovementSource {
    pub fn event_id(&self) -> &Uuid {
        match self {
            MovementSource::Command { ref event_id, .. } => event_id,
            MovementSource::Follow {
                ref origin_uuid, ..
            } => origin_uuid,
            MovementSource::Internal { ref event_id } => event_id,
        }
    }
}

#[derive(Clone, Debug, PartialEq, Eq, Ord, PartialOrd, Serialize, Deserialize)]
pub enum QueueCommand {
    CloseDoor {
        direction: Direction,
    },
    Cut {
        from_corpse: String,
        what_part: String,
    },
    Fill {
        from_item_type: String,
        from_item_code: String,
        to_item_type: String,
        to_item_code: String,
    },
    Drink {
        item_type: String,
        item_code: String,
    },
    Drop {
        possession_id: String,
    },
    Eat {
        possession_id: String,
    },
    Get {
        possession_id: String,
    },
    GetFromContainer {
        from_possession_id: String,
        get_possession_id: String,
    },
    Make {
        bench_possession_id: Option<String>,
        instructions_possession_id: String,
        already_used: BTreeSet<String>,
    },
    Movement {
        direction: Direction,
        source: MovementSource,
    },
    OpenDoor {
        direction: Direction,
    },
    Put {
        container_possession_id: String,
        target_possession_id: String,
    },
    Recline {
        item: Option<String>,
    },
    Remove {
        possession_id: String,
    },
    Scavenge,
    Sit {
        item: Option<String>,
    },
    Stand,
    Use {
        possession_id: String,
        target_id: String,
    },
    Wear {
        possession_id: String,
    },
    Wield {
        possession_id: String,
    },
    ImprovWith {
        possession_id: String,
    },
    ImprovFrom {
        output: PossessionType,
        possession_ids: BTreeSet<String>,
        already_used: BTreeSet<String>,
    },
}
impl QueueCommand {
    pub fn name(&self) -> &'static str {
        use QueueCommand::*;
        match self {
            CloseDoor { .. } => "CloseDoor",
            Cut { .. } => "Cut",
            Drink { .. } => "Drink",
            Drop { .. } => "Drop",
            Eat { .. } => "Eat",
            Fill { .. } => "Fill",
            Get { .. } => "Get",
            GetFromContainer { .. } => "GetFromContainer",
            Make { .. } => "Make",
            Movement { .. } => "Movement",
            OpenDoor { .. } => "OpenDoor",
            Put { .. } => "Put",
            Recline { .. } => "Recline",
            Remove { .. } => "Remove",
            Scavenge => "Scavenge",
            Sit { .. } => "Sit",
            Stand { .. } => "Stand",
            Use { .. } => "Use",
            Wear { .. } => "Wear",
            Wield { .. } => "Wield",
            ImprovWith { .. } => "ImprovWith",
            ImprovFrom { .. } => "ImprovFrom",
        }
    }
}

pub struct QueuedCommandContext<'l> {
    pub trans: &'l DBTrans,
    pub command: &'l QueueCommand,
    pub item: &'l mut Item,
}

impl<'l> QueuedCommandContext<'l> {
    pub async fn get_session(&self) -> UResult<Option<(ListenerSession, Session)>> {
        Ok(if self.item.item_type != "player" {
            None
        } else {
            self.trans
                .find_session_for_player(&self.item.item_code)
                .await?
        })
    }
}

#[async_trait]
pub trait QueueCommandHandler {
    async fn start_command(&self, ctx: &mut QueuedCommandContext<'_>) -> UResult<time::Duration>;
    async fn finish_command(&self, ctx: &mut QueuedCommandContext<'_>) -> UResult<()>;
}

fn queue_command_registry(
) -> &'static BTreeMap<&'static str, &'static (dyn QueueCommandHandler + Sync + Send)> {
    static REGISTRY: OnceCell<
        BTreeMap<&'static str, &'static (dyn QueueCommandHandler + Sync + Send)>,
    > = OnceCell::new();
    REGISTRY.get_or_init(|| {
        vec![
            (
                "Cut",
                &cut::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "CloseDoor",
                &close::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Drink",
                &drink::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Drop",
                &drop::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Get",
                &get::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "GetFromContainer",
                &get::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Eat",
                &eat::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Fill",
                &fill::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Make",
                &make::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Movement",
                &movement::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "OpenDoor",
                &open::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Put",
                &put::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Recline",
                &recline::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Remove",
                &remove::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Scavenge",
                &scavenge::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Sit",
                &sit::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Stand",
                &stand::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Use",
                &use_cmd::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Wear",
                &wear::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "Wield",
                &wield::QueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "ImprovWith",
                &improvise::WithQueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
            (
                "ImprovFrom",
                &improvise::FromQueueHandler as &(dyn QueueCommandHandler + Sync + Send),
            ),
        ]
        .into_iter()
        .collect()
    })
}

// Note: Saves item so you can return after calling.
pub async fn queue_command_and_save(
    ctx: &mut VerbContext<'_>,
    item: &Item,
    command: &QueueCommand,
) -> UResult<()> {
    let mut item_mut = item.clone();
    queue_command(ctx, &mut item_mut, command).await?;
    ctx.trans.save_item_model(&item_mut).await?;
    Ok(())
}

// Note: Saves item so you can return after calling.
pub async fn queue_command_for_npc_and_save(
    trans: &DBTrans,
    item: &Item,
    command: &QueueCommand,
) -> DResult<()> {
    let mut item_mut = item.clone();
    queue_command_for_npc(trans, &mut item_mut, command).await?;
    trans.save_item_model(&item_mut).await?;
    Ok(())
}

// Caller must save item or it won't actually work.
pub async fn queue_command(
    ctx: &mut VerbContext<'_>,
    item: &mut Item,
    command: &QueueCommand,
) -> UResult<()> {
    let was_empty = item.queue.is_empty();
    if item.queue.len() >= 20 {
        user_error("Can't queue more than 20 actions\n".to_owned())?;
    }
    item.queue.push_back(command.clone());
    if was_empty {
        match queue_command_registry()
            .get(&command.name())
            .expect("QueueCommand to have been registered")
            .start_command(&mut QueuedCommandContext {
                trans: ctx.trans,
                command,
                item,
            })
            .await
        {
            Err(CommandHandlingError::UserError(err_msg)) => {
                item.queue.truncate(0);
                ctx.trans
                    .queue_for_session(&ctx.session, Some(&(err_msg + "\r\n")))
                    .await?;
            }
            Err(CommandHandlingError::SystemError(e)) => Err(e)?,
            Ok(dur) => {
                ctx.trans
                    .upsert_task(&Task {
                        meta: TaskMeta {
                            task_code: item.refstr(),
                            next_scheduled: Utc::now() + chrono::Duration::from_std(dur)?,
                            ..Default::default()
                        },
                        details: TaskDetails::RunQueuedCommand,
                    })
                    .await?;
            }
        }
    } else {
        ctx.trans
            .queue_for_session(ctx.session, Some("[queued]\n"))
            .await?;
    }
    Ok(())
}

// Caller must save item or it won't actually work.
pub async fn queue_command_for_npc(
    trans: &DBTrans,
    item: &mut Item,
    command: &QueueCommand,
) -> DResult<()> {
    let was_empty = item.queue.is_empty();
    if item.queue.len() >= 20 {
        Err("Can't queue more than 20 actions\n")?;
    }
    item.queue.push_back(command.clone());
    if was_empty {
        match queue_command_registry()
            .get(&command.name())
            .expect("QueueCommand to have been registered")
            .start_command(&mut QueuedCommandContext {
                trans,
                command,
                item,
            })
            .await
        {
            Err(CommandHandlingError::UserError(_err_msg)) => {
                item.queue.truncate(0);
            }
            Err(CommandHandlingError::SystemError(e)) => Err(e)?,
            Ok(dur) => {
                trans
                    .upsert_task(&Task {
                        meta: TaskMeta {
                            task_code: item.refstr(),
                            next_scheduled: Utc::now() + chrono::Duration::from_std(dur)?,
                            ..Default::default()
                        },
                        details: TaskDetails::RunQueuedCommand,
                    })
                    .await?;
            }
        }
    }
    Ok(())
}

pub struct RunQueuedCommandTaskHandler;
#[async_trait]
impl TaskHandler for RunQueuedCommandTaskHandler {
    async fn do_task(&self, ctx: &mut TaskRunContext) -> DResult<Option<time::Duration>> {
        let (item_type, item_code) = ctx
            .task
            .meta
            .task_code
            .as_str()
            .split_once("/")
            .ok_or("QueuedCommandHandler has bad item refstr")?;
        let mut item = (*(ctx
            .trans
            .find_item_by_type_code(item_type, item_code)
            .await?
            .ok_or("Can't find player to process QueuedCommand")?))
        .clone();
        let queue_command = match item.queue.pop_front() {
            None => {
                return Ok(None);
            }
            Some(x) => x,
        };
        let mut qcontext = QueuedCommandContext {
            trans: ctx.trans,
            item: &mut item,
            command: &queue_command,
        };
        let uresult_finish = queue_command_registry()
            .get(&queue_command.name())
            .expect("QueueCommand to have been registered")
            .finish_command(&mut qcontext)
            .await;
        match uresult_finish {
            Ok(()) => {}
            Err(CommandHandlingError::UserError(err_msg)) => {
                if item.item_type == "player" {
                    if let Some((listener_sess, _)) =
                        ctx.trans.find_session_for_player(&item.item_code).await?
                    {
                        ctx.trans
                            .queue_for_session(&listener_sess, Some(&(err_msg + "\r\n")))
                            .await?;
                    }
                }
                item.queue.truncate(0);
                ctx.trans.save_item_model(&item).await?;
                return Ok(None);
            }
            Err(CommandHandlingError::SystemError(e)) => Err(e)?,
        };

        let next_command_opt = item.queue.front().cloned();
        let result = match next_command_opt {
            None => None,
            Some(next_command) => {
                let mut qcontext = QueuedCommandContext {
                    trans: ctx.trans,
                    item: &mut item,
                    command: &next_command,
                };
                match queue_command_registry()
                    .get(&next_command.name())
                    .expect("QueueCommand to have been registered")
                    .start_command(&mut qcontext)
                    .await
                {
                    Err(CommandHandlingError::UserError(err_msg)) => {
                        if item.item_type == "player" {
                            if let Some((listener_sess, _)) =
                                ctx.trans.find_session_for_player(&item.item_code).await?
                            {
                                ctx.trans
                                    .queue_for_session(&listener_sess, Some(&(err_msg + "\r\n")))
                                    .await?;
                            }
                        }
                        item.queue.truncate(0);
                        ctx.trans.save_item_model(&item).await?;
                        None
                    }
                    Err(CommandHandlingError::SystemError(e)) => Err(e)?,
                    Ok(dur) => Some(dur),
                }
            }
        };
        ctx.trans.save_item_model(&item).await?;

        Ok(result)
    }
}

pub static HANDLER: &'static (dyn TaskHandler + Sync + Send) = &RunQueuedCommandTaskHandler;
