#[double]
use crate::db::DBTrans;
use crate::{models::item::Item, static_content::possession_type::possession_data, DResult};

use mockall_double::double;

#[derive(Debug, PartialEq)]
pub enum CapacityLevel {
    Unburdened,
    SlightlyBurdened,
    HeavilyBurdened,
    OverBurdened,
    AboveItemLimit,
}

pub async fn check_item_ref_capacity(
    trans: &DBTrans,
    container: &str,
    proposed_weight: u64,
) -> DResult<CapacityLevel> {
    if let Some((item_type, item_code)) = container.split_once("/") {
        if item_type != "player" && item_type != "npc" && item_type != "possession" {
            // Fast path...
            let stats = trans.get_location_stats(&container).await?;
            if stats.total_count >= 50 || proposed_weight > 0 && stats.total_count >= 49 {
                Ok(CapacityLevel::AboveItemLimit)
            } else {
                Ok(CapacityLevel::Unburdened)
            }
        } else {
            if let Some(item) = trans.find_item_by_type_code(item_type, item_code).await? {
                check_item_capacity(trans, &item, proposed_weight).await
            } else {
                Err("Invalid item.")?
            }
        }
    } else {
        Err("Invalid item format.")?
    }
}

pub async fn check_item_capacity(
    trans: &DBTrans,
    container: &Item,
    proposed_weight: u64,
) -> DResult<CapacityLevel> {
    let stats = trans.get_location_stats(&container.refstr()).await?;
    if stats.total_count >= 50 || proposed_weight > 0 && stats.total_count >= 49 {
        return Ok(CapacityLevel::AboveItemLimit);
    }
    if container.item_type == "player"
        || container.item_type == "npc"
        || container.item_type == "possession"
    {
        let max_weight = container.max_carry();
        let new_weight = stats.total_weight + proposed_weight;
        if new_weight >= max_weight {
            return Ok(CapacityLevel::OverBurdened);
        } else if new_weight >= max_weight * 4 / 5 {
            return Ok(CapacityLevel::HeavilyBurdened);
        } else if new_weight >= max_weight / 2 {
            return Ok(CapacityLevel::SlightlyBurdened);
        }
    }
    Ok(CapacityLevel::Unburdened)
}

pub async fn recalculate_container_weight_mut(
    trans: &DBTrans,
    container: &mut Item,
) -> DResult<bool> {
    let liq_weight = match container.liquid_details.as_ref() {
        None => 0,
        Some(ld) => ld
            .contents
            .iter()
            .map(|(liq, vol)| (liq.density() * (*vol as f64)).ceil() as u64)
            .sum(),
    };

    if let Some(container_data) = container.possession_type.as_ref().and_then(|pt| {
        possession_data()
            .get(pt)
            .and_then(|pd| pd.container_data.as_ref())
    }) {
        let stats = trans.get_location_stats(&container.refstr()).await?;
        let new_weight = container_data.base_weight
            + (((stats.total_weight as f64) * container_data.compression_ratio).ceil() as u64)
            + liq_weight;
        if new_weight != container.weight {
            container.weight = new_weight;
            Ok(true)
        } else {
            Ok(false)
        }
    } else if liq_weight > 0 {
        if let Some(pd) = container
            .possession_type
            .as_ref()
            .and_then(|pt| possession_data().get(pt))
        {
            let new_weight = liq_weight + pd.weight;
            if new_weight != container.weight {
                container.weight = new_weight;
                Ok(true)
            } else {
                Ok(false)
            }
        } else {
            Ok(false)
        }
    } else {
        Ok(false)
    }
}

pub async fn recalculate_container_weight(trans: &DBTrans, container: &Item) -> DResult<()> {
    let mut container_mut = (*container).clone();
    if recalculate_container_weight_mut(trans, &mut container_mut).await? {
        trans.save_item_model(&container_mut).await?;
    }
    Ok(())
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::{
        db::{LocationStats, MockDBTrans},
        models::item::StatType,
    };

    #[tokio::test]
    async fn check_item_capacity_should_say_above_item_limit_if_over() {
        let mut mock_db = MockDBTrans::new();
        mock_db
            .expect_find_item_by_type_code()
            .withf(|t, c| t == "player" && c == "foo")
            .returning(|_, _| {
                Ok(Some(
                    Item {
                        item_type: "player".to_owned(),
                        item_code: "foo".to_owned(),
                        ..Default::default()
                    }
                    .into(),
                ))
            });
        mock_db
            .expect_get_location_stats()
            .withf(|s| s == "player/foo")
            .returning(|_| {
                Ok(LocationStats {
                    total_count: 49,
                    total_weight: 100,
                })
            });
        assert_eq!(
            check_item_ref_capacity(&mock_db, "player/foo", 10)
                .await
                .unwrap(),
            CapacityLevel::AboveItemLimit
        );
    }

    #[tokio::test]
    async fn check_item_capacity_should_say_overburdened_if_over() {
        let mut mock_db = MockDBTrans::new();
        mock_db
            .expect_find_item_by_type_code()
            .withf(|t, c| t == "player" && c == "foo")
            .returning(|_, _| {
                Ok(Some(
                    Item {
                        item_type: "player".to_owned(),
                        item_code: "foo".to_owned(),
                        ..Default::default()
                    }
                    .into(),
                ))
            });
        mock_db
            .expect_get_location_stats()
            .withf(|s| s == "player/foo")
            .returning(|_| {
                Ok(LocationStats {
                    total_count: 2,
                    total_weight: 100,
                })
            });
        assert_eq!(
            check_item_ref_capacity(&mock_db, "player/foo", 1000000)
                .await
                .unwrap(),
            CapacityLevel::OverBurdened
        );
    }

    #[tokio::test]
    async fn check_item_capacity_should_say_unburdened_when_low_weight() {
        let mut mock_db = MockDBTrans::new();
        mock_db
            .expect_find_item_by_type_code()
            .withf(|t, c| t == "player" && c == "foo")
            .returning(|_, _| {
                Ok(Some(
                    Item {
                        item_type: "player".to_owned(),
                        item_code: "foo".to_owned(),
                        total_stats: vec![(StatType::Brawn, 8.0)].into_iter().collect(),
                        ..Default::default()
                    }
                    .into(),
                ))
            });
        mock_db
            .expect_get_location_stats()
            .withf(|s| s == "player/foo")
            .returning(|_| {
                Ok(LocationStats {
                    total_count: 2,
                    total_weight: 100,
                })
            });
        assert_eq!(
            check_item_ref_capacity(&mock_db, "player/foo", 50)
                .await
                .unwrap(),
            CapacityLevel::Unburdened
        );
    }
}
