#[double]
use crate::db::DBTrans;
use crate::{
    message_handler::user_commands::rent::recursively_destroy_or_move_item,
    models::{
        item::{LiquidDetails, LiquidType, LocationActionType},
        task::{Task, TaskDetails, TaskMeta, TaskRecurrence},
    },
    regular_tasks::{TaskHandler, TaskRunContext},
    services::Item,
    static_content::{
        possession_type::PossessionType, room::room_list, scavtable::scavtable_map, StaticTask,
    },
    DResult,
};
use async_recursion::async_recursion;
use async_trait::async_trait;
use log::{info, warn};
use mockall_double::double;
use once_cell::sync::OnceCell;
use rand::{thread_rng, Rng};
use rand_distr::Distribution;
use std::collections::BTreeMap;
use std::time;

#[allow(dead_code)]
pub enum SpawnDistribution {
    DoNothing,
    SpawnLiquid {
        what: LiquidType,
        how_much: u64,
    },
    SpawnPossession {
        what: PossessionType,
    },
    SpawnWithProb {
        pspawn: f64,
        subspawn: Box<SpawnDistribution>,
    },
    SpawnOne {
        pvec: Vec<(f64, Box<SpawnDistribution>)>,
    },
    SpawnAll {
        subspawns: Vec<Box<SpawnDistribution>>,
    },
}

#[async_recursion]
async fn add_per_spawn(
    trans: &DBTrans,
    distribution: &SpawnDistribution,
    location: &mut Item,
) -> DResult<bool> {
    match distribution {
        SpawnDistribution::DoNothing => Ok(false),
        SpawnDistribution::SpawnLiquid { what, how_much } => {
            match location.liquid_details.as_mut() {
                None => {
                    location.liquid_details = Some(LiquidDetails {
                        contents: vec![(what.clone(), how_much.clone())].into_iter().collect(),
                        ..Default::default()
                    });
                    Ok(true)
                }
                Some(liq_dets) => {
                    liq_dets
                        .contents
                        .entry(what.clone())
                        .and_modify(|d| *d += how_much)
                        .or_insert(how_much.clone());
                    Ok(true)
                }
            }
        }
        SpawnDistribution::SpawnPossession { what } => {
            let mut item: Item = what.clone().into();
            item.location = location.refstr();
            trans.create_item(&item).await?;
            Ok(false)
        }
        SpawnDistribution::SpawnWithProb { pspawn, subspawn } => {
            let samp: f64 = thread_rng().gen();
            if samp < *pspawn {
                add_per_spawn(trans, &subspawn, location).await
            } else {
                Ok(false)
            }
        }
        SpawnDistribution::SpawnOne { pvec } => {
            let samp: f64 = thread_rng().gen();
            let mut cumsum = 0.0_f64;
            for opt in pvec {
                cumsum += opt.0;
                if samp < cumsum {
                    return add_per_spawn(trans, &opt.1, location).await;
                }
            }
            Ok(false)
        }
        SpawnDistribution::SpawnAll { subspawns } => {
            let mut did_mut_location = false;
            for spawn in subspawns {
                did_mut_location |= add_per_spawn(trans, &spawn, location).await?
            }
            Ok(did_mut_location)
        }
    }
}

fn spawn_list() -> &'static BTreeMap<&'static str, SpawnDistribution> {
    static SPAWN_LIST: OnceCell<BTreeMap<&'static str, SpawnDistribution>> = OnceCell::new();
    SPAWN_LIST.get_or_init(|| {
        vec![(
            "fixed_item/melbs_king_st_spring_fed_fountain",
            SpawnDistribution::SpawnOne {
                pvec: vec![(
                    1.0,
                    Box::new(SpawnDistribution::SpawnLiquid {
                        what: LiquidType::Water,
                        how_much: 1000000,
                    }),
                )],
            },
        )]
        .into_iter()
        .collect()
    })
}

pub async fn refresh_all_spawn_points(trans: &DBTrans) -> DResult<()> {
    for (location_ref, distro) in spawn_list().iter() {
        let (loc_type, loc_code) = match location_ref.split_once("/") {
            None => {
                warn!("Invalid spawn point location {}", location_ref);
                continue;
            }
            Some(v) => v,
        };
        let location = match trans.find_item_by_type_code(loc_type, loc_code).await? {
            None => {
                warn!("Spawn point location {} not found", location_ref);
                continue;
            }
            Some(v) => v,
        };
        // Step 1: Clear spawned site. It's paginated so we loop to get everything...
        loop {
            let result = trans.find_items_by_location(location_ref).await?;
            if result.is_empty() {
                break;
            }
            for sub_item in result {
                recursively_destroy_or_move_item(trans, &sub_item).await?;
            }
        }
        let mut location_mut = (*location).clone();
        let mut need_save = false;
        match location_mut.liquid_details.as_mut() {
            None => {}
            Some(liq_dets) => {
                if !liq_dets.contents.is_empty() {
                    need_save = true;
                    liq_dets.contents = BTreeMap::new();
                }
            }
        }
        // Step 2: Add any spawns that come up
        need_save |= add_per_spawn(trans, distro, &mut location_mut).await?;
        // Step 3: Save the location if needed (for liquids)
        if need_save {
            trans.save_item_model(&location_mut).await?;
        }
    }

    // Also all scav spawns...
    trans.clean_scavhidden().await?;
    let empty_vec = vec![];
    for room in room_list().iter() {
        for scav in scavtable_map()
            .get(&room.scavtable)
            .unwrap_or_else(|| &empty_vec)
        {
            if thread_rng().gen_bool(scav.p_present) {
                let difflevel = {
                    let mut rng = thread_rng();
                    (rand_distr::Normal::new(scav.difficulty_mean, scav.difficulty_stdev)?
                        .sample(&mut rng)
                        * 100.0)
                        .max(0.0) as u64
                };
                let mut item: Item = scav.possession_type.clone().into();
                item.item_code = format!("{}", trans.alloc_item_code().await?);
                item.location = format!("room/{}", room.code);
                item.action_type = LocationActionType::Scavhidden {
                    difficulty: difflevel,
                    scavtype: scav.scavtype.clone(),
                };
                trans.save_item_model(&item).await?;
            }
        }
    }

    Ok(())
}

pub struct ResetSpawnsTaskHandler;
#[async_trait]
impl TaskHandler for ResetSpawnsTaskHandler {
    async fn do_task(&self, ctx: &mut TaskRunContext) -> DResult<Option<time::Duration>> {
        info!("Running spawn point refresh (timed)");
        refresh_all_spawn_points(ctx.trans).await?;
        Ok(Some(time::Duration::from_secs(3600)))
    }
}
pub static RESET_SPAWNS_HANDLER: &'static (dyn TaskHandler + Sync + Send) = &ResetSpawnsTaskHandler;

pub fn reset_tasks() -> Box<dyn Iterator<Item = StaticTask>> {
    Box::new(
        vec![StaticTask {
            task_code: "ResetSpawns".to_owned(),
            initial_task: Box::new(|| Task {
                meta: TaskMeta {
                    task_code: "ResetSpawns".to_owned(),
                    is_static: true,
                    recurrence: Some(TaskRecurrence::FixedDuration { seconds: 3600 }),
                    ..Default::default()
                },
                details: TaskDetails::ResetSpawns,
            }),
        }]
        .into_iter(),
    )
}
