use super::{super::room::GridCoords, Dynroom, Dynzone, DynzoneType, Exit, ExitTarget, ExitType};
use crate::models::item::ItemFlag;
use crate::static_content::room::Direction;

pub fn zone() -> Dynzone {
    Dynzone {
        zonetype: DynzoneType::MurlDeluxeCorporate,
        zonename: "Murlison Suites",
        entrypoint_subcode: "reception",
        dyn_rooms: vec!(
            ("reception", Dynroom {
                subcode: "reception",
                name: "Reception",
                short: "RC",
                description: "A narrow public area of the corporate suite, with views \
                              of the bleak wasteland to the north. The walls are painted white, and a long reception \
                              desk, enamelled to look like dark wood, is built into the room. It is well lit \
                              with cool white LED lights recessed into the ceiling above",
                exits: vec!(
                    Exit {
                        direction: Direction::WEST,
                        target: ExitTarget::ExitZone,
                        exit_type: ExitType::Doorless
                    },
                    Exit {
                        direction: Direction::EAST,
                        target: ExitTarget::Intrazone { subcode: "commons" },
                        exit_type: ExitType::Doorless
                    }
                ),
                grid_coords: GridCoords { x: 0, y: 0, z: 0 },
                should_caption: true,
                item_flags: vec!(),
                ..Default::default()
            }),
            ("commons", Dynroom {
                subcode: "commons",
                name: "Common Room",
                short: "CO",
                description: "A large central room, carpeted with floor tiles alternating between light and dark grey in a checkerboard pattern. The walls are painted white, and the room is dimly lit from above with warm white LED bulbs",
                exits: vec!(
                    Exit {
                        direction: Direction::WEST,
                        target: ExitTarget::Intrazone { subcode: "reception" },
                        exit_type: ExitType::Doored { description: "a very sturdy looking fire-rated solid core navy blue painted door" }
                    },
                    Exit {
                        direction: Direction::NORTH,
                        target: ExitTarget::Intrazone { subcode: "northwing" },
                        exit_type: ExitType::Doorless
                    },
                    Exit {
                        direction: Direction::EAST,
                        target: ExitTarget::Intrazone { subcode: "eastwing" },
                        exit_type: ExitType::Doorless
                    },
                    Exit {
                        direction: Direction::SOUTHWEST,
                        target: ExitTarget::Intrazone { subcode: "healthcentre" },
                        exit_type: ExitType::Doorless
                    },
                    Exit {
                        direction: Direction::SOUTH,
                        target: ExitTarget::Intrazone { subcode: "kitchen" },
                        exit_type: ExitType::Doorless
                    },
                ),
                grid_coords: GridCoords { x: 1, y: 0, z: 0 },
                should_caption: true,
                has_power: true,
                item_flags: vec!(ItemFlag::DroppedItemsDontExpire,
                                 ItemFlag::PrivatePlace),
                ..Default::default()
            }),
            ("northwing", Dynroom {
                subcode: "northwing",
                name: "North Wing",
                short: "NW",
                description: "A spacious office, carpeted with floor tiles alternating between light and dark grey in a checkerboard pattern. The walls are painted white, and the room is dimly lit from above with warm white LED bulbs",
                exits: vec!(
                    Exit {
                        direction: Direction::NORTH,
                        target: ExitTarget::Intrazone { subcode: "northbalcony" },
                        exit_type: ExitType::Doorless
                    },
                    Exit {
                        direction: Direction::SOUTH,
                        target: ExitTarget::Intrazone { subcode: "commons" },
                        exit_type: ExitType::Doorless
                    },
                ),
                grid_coords: GridCoords { x: 1, y: -1, z: 0 },
                should_caption: true,
                has_power: true,
                item_flags: vec!(ItemFlag::DroppedItemsDontExpire,
                                 ItemFlag::PrivatePlace),
                ..Default::default()
            }),
            ("northbalcony", Dynroom {
                subcode: "northbalcony",
                name: "North Balcony",
                short: "NB",
                description: "A timber balcony cantilevered on to the north side of the building, complete with an auto-irrigated garden, and glass front panels exposing a breath-taking view of the desolate lands to the north",
                exits: vec!(
                    Exit {
                        direction: Direction::SOUTH,
                        target: ExitTarget::Intrazone { subcode: "commons" },
                        exit_type: ExitType::Doored { description: "a pair of sliding glass doors" }
                    },
                ),
                grid_coords: GridCoords { x: 1, y: -2, z: 0 },
                should_caption: true,
                has_power: false,
                item_flags: vec!(ItemFlag::DroppedItemsDontExpire,
                                 ItemFlag::PrivatePlace),
                ..Default::default()
            }),
            ("eastwing", Dynroom {
                subcode: "eastwing",
                name: "East Wing",
                short: "EW",
                description: "A spacious office, carpeted with floor tiles alternating between light and dark grey in a checkerboard pattern. The walls are painted white, and the room is dimly lit from above with warm white LED bulbs",
                exits: vec!(
                    Exit {
                        direction: Direction::EAST,
                        target: ExitTarget::Intrazone { subcode: "storeroom" },
                        exit_type: ExitType::Doorless
                    },
                    Exit {
                        direction: Direction::WEST,
                        target: ExitTarget::Intrazone { subcode: "commons" },
                        exit_type: ExitType::Doorless
                    },
                ),
                grid_coords: GridCoords { x: 2, y: 0, z: 0 },
                should_caption: true,
                has_power: true,
                item_flags: vec!(ItemFlag::DroppedItemsDontExpire,
                                 ItemFlag::PrivatePlace),
                ..Default::default()
            }),
            ("storeroom", Dynroom {
                subcode: "storeroom",
                name: "Store room",
                short: "SR",
                description: "A strong-room apparently designed to preserve people or things in any emergency. The walls, floor, and ceiling all appear to be made of thick steel",
                exits: vec!(
                    Exit {
                        direction: Direction::WEST,
                        target: ExitTarget::Intrazone { subcode: "eastwing" },
                        exit_type: ExitType::Doored { description: "a heavy thick-plated steel door" }
                    },
                ),
                grid_coords: GridCoords { x: 3, y: 0, z: 0 },
                should_caption: true,
                has_power: true,
                item_flags: vec!(ItemFlag::DroppedItemsDontExpire,
                                 ItemFlag::PrivatePlace),
                ..Default::default()
            }),
            ("healthcentre", Dynroom {
                subcode: "healthcentre",
                name: "Health centre",
                short: "HC",
                description: "A room apparently designed as a kind of corporate sick-bay. The floor is tiled, and the walls are painted with white glossy easy-wipe paint",
                exits: vec!(
                    Exit {
                        direction: Direction::NORTHEAST,
                        target: ExitTarget::Intrazone { subcode: "commons" },
                        exit_type: ExitType::Doorless
                    },
                ),
                grid_coords: GridCoords { x: 0, y: 1, z: 0 },
                should_caption: true,
                has_power: true,
                item_flags: vec!(ItemFlag::DroppedItemsDontExpire,
                                 ItemFlag::PrivatePlace),
                ..Default::default()
            }),
            ("kitchen", Dynroom {
                subcode: "kitchen",
                name: "Kitchen",
                short: "KT",
                description: "A room apparently designed for food preparation. It is tiled with small light green tiles that cover the walls and floor, and has an abundance of water and drain fittings",
                exits: vec!(
                    Exit {
                        direction: Direction::NORTH,
                        target: ExitTarget::Intrazone { subcode: "commons" },
                        exit_type: ExitType::Doorless
                    },
                ),
                grid_coords: GridCoords { x: 1, y: 1, z: 0 },
                should_caption: true,
                has_power: true,
                item_flags: vec!(ItemFlag::DroppedItemsDontExpire,
                                 ItemFlag::PrivatePlace),
                ..Default::default()
            }),
        ).into_iter().collect(),
        ..Default::default()
    }
}
