use super::{PossessionData, PossessionType};
use crate::{
    message_handler::user_commands::{UResult, UserError},
    models::item::{Item, ItemFlag},
    static_content::possession_type::{ContainerCheck, ContainerData},
};
use once_cell::sync::OnceCell;
use std::collections::BTreeSet;

pub fn recipe_set() -> &'static BTreeSet<PossessionType> {
    static SET: OnceCell<BTreeSet<PossessionType>> = OnceCell::new();
    &SET.get_or_init(|| {
        vec![PossessionType::GrilledSteakRecipe]
            .into_iter()
            .collect()
    })
}

struct RecipesOnlyChecker;
impl ContainerCheck for RecipesOnlyChecker {
    fn check_place(&self, _container: &Item, item: &Item) -> UResult<()> {
        item.possession_type
            .as_ref()
            .and_then(|pt| recipe_set().get(pt))
            .ok_or_else(|| {
                UserError("You don't find a sensible place for that in a recipe book.".to_owned())
            })?;
        Ok(())
    }
}

static RECIPES_ONLY_CHECKER: RecipesOnlyChecker = RecipesOnlyChecker;

pub fn data() -> &'static Vec<(PossessionType, PossessionData)> {
    static D: OnceCell<Vec<(PossessionType, PossessionData)>> = OnceCell::new();
    &D.get_or_init(|| vec!(
        (PossessionType::CulinaryEssentials,
         PossessionData {
             display: "Culinary Essentials cookbook",
             aliases: vec!["book", "cookbook"],
             details: "A weathered cookbook filled with essential recipes for survival in the post-apocalyptic world. Its pages are yellowed and marked with stains, a testament to its extensive use by those seeking sustenance and comfort amidst scarcity.",
             weight: 300,
             container_data: Some(ContainerData {
                 base_weight: 290,
                 default_contents: vec![
                     PossessionType::GrilledSteakRecipe
                 ],
                 checker: &RECIPES_ONLY_CHECKER,
                 ..Default::default()
             }),
             default_flags: vec![ItemFlag::Book],
             ..Default::default()
         }),
        (PossessionType::GrilledSteakRecipe,
         PossessionData {
             display: "recipe for Grilled Steak",
             aliases: vec!["grilled steak"],
             details: "Instructions for how to make a basic but mouthwatering steak",
             weight: 10,
             default_flags: vec![ItemFlag::Instructions],
             ..Default::default()
         }),
    ))
}
