use super::{possession_data, ArglessHandler, PossessionData, PossessionType, WriteHandler};
use crate::{
    message_handler::user_commands::{
        parsing::parse_username, register::is_invalid_username, user_error,
        CommandHandlingError::UserError, UResult, VerbContext,
    },
    models::{
        corp::{Corp, CorpMembership, CorpPermission},
        item::{Item, ItemSpecialData},
    },
    services::comms::broadcast_to_room,
};
use ansi::ansi;
use async_trait::async_trait;
use chrono::Utc;
use once_cell::sync::OnceCell;

use super::PossessionType::*;

pub struct CorpLicenceHandler {}

#[async_trait]
impl WriteHandler for CorpLicenceHandler {
    async fn write_cmd(
        &self,
        ctx: &mut VerbContext,
        _player: &Item,
        on_what: &Item,
        write_what: &str,
    ) -> UResult<()> {
        let name = match parse_username(write_what) {
            Err(e) => user_error("Invalid corp name: ".to_owned() + e)?,
            Ok((_, rest)) if rest != "" => {
                user_error("No spaces allowed in corp names!".to_owned())?
            }
            Ok((name, _)) => name,
        };
        if is_invalid_username(name) {
            user_error("Sorry, that corp name isn't allowed. Try another".to_owned())?;
        }
        if ctx.trans.find_by_username(name).await?.is_some() {
            user_error("Corp name clashes with existing user name".to_owned())?;
        }
        if ctx.trans.find_corp_by_name(&name).await?.is_some() {
            user_error("Corp name already taken!".to_owned())?;
        }

        let mut item_clone = on_what.clone();
        item_clone.special_data = Some(ItemSpecialData::ItemWriting {
            text: name.to_owned(),
        });
        ctx.trans.save_item_model(&item_clone).await?;
        ctx.trans
            .queue_for_session(
                ctx.session,
                Some(&format!(
                    ansi!(
            "The pencil makes a scratching sound as you mark the paper with the attached \
            pencil and write \"{}\" on it. [Hint: Try the <bold>use<reset> command to submit \
            your signed paperwork and register the corporation, or <bold>write<reset> again \
            to erase and change the name].\n"),
                    name
                )),
            )
            .await?;

        Ok(())
    }
}

#[async_trait]
impl ArglessHandler for CorpLicenceHandler {
    async fn cmd(&self, ctx: &mut VerbContext, player: &Item, what: &Item) -> UResult<()> {
        let name = match what.special_data.as_ref() {
            Some(ItemSpecialData::ItemWriting { text }) => text,
            _ => user_error("You have to write your corp's name on it first!".to_owned())?,
        };
        if ctx.trans.find_by_username(&name).await?.is_some() {
            user_error("Corp name clashes with existing user name".to_owned())?;
        }
        if ctx.trans.find_corp_by_name(&name).await?.is_some() {
            user_error("Corp name already taken!".to_owned())?;
        }

        if ctx
            .trans
            .get_corp_memberships_for_user(&player.item_code)
            .await?
            .len()
            >= 5
        {
            user_error("You can't be in more than 5 corps".to_owned())?;
        }

        broadcast_to_room(
            ctx.trans,
            &player.location,
            None,
            &format!(
                "{} signs a contract establishing {} as a corp\n",
                &player.display_for_sentence(1, true),
                name
            ),
        )
        .await?;
        let corp_id = ctx
            .trans
            .create_corp(&Corp {
                name: name.to_owned(),
                ..Default::default()
            })
            .await?;
        ctx.trans
            .upsert_corp_membership(
                &corp_id,
                &player.item_code,
                &CorpMembership {
                    joined_at: Some(Utc::now()),
                    permissions: vec![CorpPermission::Holder],
                    allow_combat: true,
                    job_title: "Founder".to_owned(),
                    ..Default::default()
                },
            )
            .await?;

        let mut what_mut = what.clone();
        what_mut.possession_type = Some(CertificateOfIncorporation);
        let cp_data = possession_data()
            .get(&CertificateOfIncorporation)
            .ok_or_else(|| {
                UserError("Certificate of Incorporation no longer exists as an item".to_owned())
            })?;
        what_mut.display = cp_data.display.to_owned();
        what_mut.details = Some(cp_data.details.to_owned());
        ctx.trans.save_item_model(&what_mut).await?;

        Ok(())
    }
}

static CORP_LICENCE_HANDLER: CorpLicenceHandler = CorpLicenceHandler {};

pub fn data() -> &'static Vec<(PossessionType, PossessionData)> {
    static D: OnceCell<Vec<(PossessionType, PossessionData)>> = OnceCell::new();
    &D.get_or_init(|| vec!(
        (NewCorpLicence,
         PossessionData {
             display: "new corp licence",
             details: ansi!("A blank form that you can <bold>use<reset> to establish a new corp. It rests on a clipboard with a pencil attached by a chain. There is a space to <bold>write<reset> on it [try <bold>write Blah on licence<reset> followed by <bold>sign licence<reset> to create a corp named Blah]"),
             aliases: vec!("form", "license", "licence", "new"),
             weight: 10,
             becomes_on_spent: Some(CertificateOfIncorporation),
             write_handler: Some(&CORP_LICENCE_HANDLER),
             sign_handler: Some(&CORP_LICENCE_HANDLER),
             ..Default::default()
         }),
        (
            CertificateOfIncorporation,
            PossessionData {
                display: "certificate of incorporation",
                details: "A certificate recording the formation of a corp.",
                weight: 10,
                ..Default::default()
            }
        ),
    ))
}
