use super::{LockcheckHandler, PossessionData, PossessionType};
use crate::{
    message_handler::user_commands::{user_error, UResult, VerbContext},
    models::item::{Item, LocationActionType},
    regular_tasks::queued_command::QueuedCommandContext,
    services::{
        capacity::{check_item_capacity, CapacityLevel},
        comms::broadcast_to_room,
    },
    static_content::{possession_type::InstallHandler, room::Direction},
};
use async_trait::async_trait;
use once_cell::sync::OnceCell;

struct BasicLockLockcheck;
#[async_trait]
impl LockcheckHandler for BasicLockLockcheck {
    async fn cmd(&self, _ctx: &mut QueuedCommandContext, _what: &Item) -> UResult<()> {
        // TODO implement actual check for key once we have them.
        user_error(
            "You search your possessions for a key that fits the door, but end up just rattling the handle uselessly.".to_owned())?;
        Ok(())
    }
}

struct ScanLockLockcheck;
#[async_trait]
impl LockcheckHandler for ScanLockLockcheck {
    async fn cmd(&self, ctx: &mut QueuedCommandContext, what: &Item) -> UResult<()> {
        if what.owner == Some(ctx.item.refstr()) {
            if let Some((session, _)) = ctx.get_session().await? {
                ctx.trans.queue_for_session(&session, Some(
                    "The scanlock in the door emits a single high-pitched beep as it unlocks.\n")).await?;
            }
        } else {
            user_error(
                "The scanlock in the door emits a medium-pitched tone followed by a low tone, and remains locked.".to_owned())?;
        }
        Ok(())
    }
}

static LOCK_WEIGHT: u64 = 500;

struct ScanLockInstall;
#[async_trait]
impl InstallHandler for ScanLockInstall {
    async fn install_cmd(
        &self,
        ctx: &mut VerbContext,
        player: &Item,
        what: &Item,
        room: &Item,
        direction: &Direction,
    ) -> UResult<()> {
        if what.action_type != LocationActionType::Normal {
            user_error("That scanlock is already in use.".to_owned())?;
        }

        if !ctx
            .trans
            .find_by_action_and_location(
                &room.refstr(),
                &LocationActionType::InstalledOnDoorAsLock(direction.clone()),
            )
            .await?
            .is_empty()
        {
            user_error("There's already a lock on that door - uninstall it first.".to_owned())?;
        }

        match check_item_capacity(&ctx.trans, &room, LOCK_WEIGHT).await? {
            CapacityLevel::OverBurdened | CapacityLevel::AboveItemLimit => user_error(
                "That room has so much stuff, you can't install anything new.".to_owned(),
            )?,
            _ => {}
        };
        let mut what_mut = (*what).clone();
        what_mut.location = room.refstr();
        what_mut.action_type = LocationActionType::InstalledOnDoorAsLock(direction.clone());
        what_mut.owner = room.owner.clone();
        ctx.trans.save_item_model(&what_mut).await?;

        broadcast_to_room(
            &ctx.trans,
            &room.refstr(),
            None,
            &format!(
                "{} bangs the door to the {} as he installs {} on it.\n",
                &player.display_for_sentence(1, true),
                &direction.describe(),
                &what.display_for_sentence(1, false)
            ),
        )
        .await?;
        Ok(())
    }

    async fn uninstall_cmd(
        &self,
        ctx: &mut VerbContext,
        player: &Item,
        what: &Item,
        room: &Item,
        direction: &Direction,
    ) -> UResult<()> {
        if what.action_type != LocationActionType::InstalledOnDoorAsLock(direction.clone()) {
            user_error("That scanlock is not installed as a lock.".to_owned())?;
        }

        let mut what_mut = (*what).clone();

        let extra_text = match check_item_capacity(&ctx.trans, &player, LOCK_WEIGHT).await? {
            CapacityLevel::OverBurdened | CapacityLevel::AboveItemLimit => {
                ", dropping it on the floor since he can't hold it."
            }
            _ => {
                what_mut.location = player.refstr();
                ""
            }
        };
        what_mut.action_type = LocationActionType::Normal;
        what_mut.owner = None;
        ctx.trans.save_item_model(&what_mut).await?;

        broadcast_to_room(
            &ctx.trans,
            &room.refstr(),
            None,
            &format!(
                "{} bangs the door to the {} as he uninstalls {} from it{}.\n",
                &player.display_for_sentence(1, true),
                &direction.describe(),
                &what.display_for_sentence(1, false),
                extra_text
            ),
        )
        .await?;
        Ok(())
    }
}

pub fn data() -> &'static Vec<(PossessionType, PossessionData)> {
    static D: OnceCell<Vec<(PossessionType, PossessionData)>> = OnceCell::new();
    &D.get_or_init(|| vec!(
        (PossessionType::Basiclock,
         PossessionData {
             display: "basic lock",
             details: "A basic lock that looks like it needs a key to open.",
             aliases: vec!("lock"),
             weight: LOCK_WEIGHT,
             lockcheck_handler: Some(&BasicLockLockcheck),
             install_handler: None,
             ..Default::default()
         }
        ),
        (PossessionType::Scanlock,
         PossessionData {
             display: "scanlock",
             details: "A relatively basic lock with a fingerprint scanner built into it, made to ensure only the owner of something can enter or use it.",
             aliases: vec!("lock"),
             weight: LOCK_WEIGHT,
             lockcheck_handler: Some(&ScanLockLockcheck),
             install_handler: Some(&ScanLockInstall),
             ..Default::default()
         }
        )
    ))
}
